from itertools import chain
from typing import List, Tuple

import pandas as pd
import torch
import torch.nn.functional as F
from tqdm import tqdm

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.nodes.passagereranker.tart.modeling_enc_t5 import EncT5ForSequenceClassification
from autorag.nodes.passagereranker.tart.tokenization_enc_t5 import EncT5Tokenizer
from autorag.utils.util import make_batch, sort_by_scores, flatten_apply, select_top_k


@passage_reranker_node
def tart(queries: List[str], contents_list: List[List[str]],
         scores_list: List[List[float]], ids_list: List[List[str]],
         top_k: int, instruction: str = "Find passage to answer given question",
         batch: int = 64) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using Tart.
    TART is a reranker based on TART (https://github.com/facebookresearch/tart).
    You can rerank the passages with the instruction using TARTReranker.
    The default model is facebook/tart-full-flan-t5-xl.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param instruction: The instruction for reranking.
        Note: default instruction is "Find passage to answer given question"
            The default instruction from the TART paper is being used.
            If you want to use a different instruction, you can change the instruction through this parameter
    :param batch: The number of queries to be processed in a batch
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model_name = "facebook/tart-full-flan-t5-xl"
    model = EncT5ForSequenceClassification.from_pretrained(model_name)
    tokenizer = EncT5Tokenizer.from_pretrained(model_name)
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = model.to(device)

    nested_list = [[['{} [SEP] {}'.format(instruction, query)] for _ in contents] for query, contents in
                   zip(queries, contents_list)]

    rerank_scores = flatten_apply(tart_run_model, nested_list, model=model, batch_size=batch,
                                  tokenizer=tokenizer, device=device, contents_list=contents_list)

    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': rerank_scores,
    })
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    del model
    del tokenizer
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


def tart_run_model(input_texts, contents_list, model, batch_size: int, tokenizer, device):
    flattened_texts = list(chain.from_iterable(input_texts))
    flattened_contents = list(chain.from_iterable(contents_list))
    batch_input_texts = make_batch(flattened_texts, batch_size)
    batch_contents_list = make_batch(flattened_contents, batch_size)
    results = []
    for batch_texts, batch_contents in tqdm(zip(batch_input_texts, batch_contents_list)):
        feature = tokenizer(batch_texts, batch_contents, padding=True, truncation=True,
                            return_tensors="pt").to(device)
        with torch.no_grad():
            pred_scores = model(**feature).logits
            normalized_scores = [float(score[1]) for score in F.softmax(pred_scores, dim=1)]
        results.extend(normalized_scores)
    return results



import asyncio
import os
from typing import List, Tuple, Optional

import cohere
from cohere import RerankResponseResultsItem

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import process_batch


@passage_reranker_node
def cohere_reranker(queries: List[str], contents_list: List[List[str]],
                    scores_list: List[List[float]], ids_list: List[List[str]],
                    top_k: int, batch: int = 64,
                    model: str = "rerank-multilingual-v2.0",
                    api_key: Optional[str] = None) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents with Cohere rerank models.
    You can get the API key from https://cohere.com/rerank and set it in the environment variable COHERE_API_KEY.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
    :param model: The model name for Cohere rerank.
        You can choose between "rerank-multilingual-v2.0" and "rerank-english-v2.0".
        Default is "rerank-multilingual-v2.0".
    :param api_key: The API key for Cohere rerank.
        You can set it in the environment variable COHERE_API_KEY.
        Or, you can directly set it on the config YAML file using this parameter.
        Default is env variable "COHERE_API_KEY".
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    api_key = os.getenv("COHERE_API_KEY", None) if api_key is None else api_key
    if api_key is None:
        raise KeyError("Please set the API key for Cohere rerank in the environment variable COHERE_API_KEY "
                       "or directly set it on the config YAML file.")

    cohere_client = cohere.AsyncClient(api_key)

    # Run async cohere_rerank_pure function
    tasks = [cohere_rerank_pure(cohere_client, model, query, document, ids, top_k) for query, document, ids in
             zip(queries, contents_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del cohere_client

    return content_result, id_result, score_result


async def cohere_rerank_pure(cohere_client: cohere.AsyncClient, model: str,
                             query: str, documents: List[str], ids: List[str],
                             top_k: int) -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents with Cohere rerank models.

    :param cohere_client: The Cohere AsyncClient to use for reranking
    :param model: The model name for Cohere rerank
    :param query: The query to use for reranking
    :param documents: The list of contents to rerank
    :param ids: The list of ids corresponding to the documents
    :param top_k: The number of passages to be retrieved
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    rerank_results = await cohere_client.rerank(model=model, query=query, documents=documents, top_n=top_k,
                                                return_documents=False)
    results: List[RerankResponseResultsItem] = rerank_results.results
    reranked_scores: List[float] = list(map(lambda x: x.relevance_score, results))
    indices = list(map(lambda x: x.index, results))
    reranked_contents: List[str] = list(map(lambda i: documents[i], indices))
    reranked_ids: List[str] = list(map(lambda i: ids[i], indices))
    return reranked_contents, reranked_ids, reranked_scores



import asyncio
import functools
import itertools
import os
from typing import List, Optional

import evaluate
import pandas as pd
from sacrebleu.metrics.bleu import BLEU
import torch
from llama_index.core.embeddings import BaseEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding
from openai import AsyncOpenAI
from rouge_score import tokenizers
from rouge_score.rouge_scorer import RougeScorer

from autorag import embedding_models
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.utils.util import process_batch, openai_truncate_by_token, convert_inputs_to_list


def generation_metric(func):
    @functools.wraps(func)
    @convert_inputs_to_list
    def wrapper(generation_gt: List[List[str]], generations: List[str], **kwargs) -> List[float]:
        """
        Compute generation metric.

        :param generation_gt: A list of ground truth.
            Must be 2-d list of string.
            Because it can be a multiple ground truth.
        :param generations: A list of generations that LLM generated.
        :param kwargs: The additional arguments for metric function.
        :return: A list of computed metric scores.
        """
        # make generation_gt and generations to pd dataframe
        result = list(map(lambda x: func(x[0], x[1], **kwargs), zip(generation_gt, generations)))
        return result

    return wrapper


@convert_inputs_to_list
def huggingface_evaluate(instance, key: str,
                         generation_gt: List[List[str]], generations: List[str],
                         **kwargs) -> List[float]:
    """
    Compute huggingface evaluate metric.

    :param instance: The instance of huggingface evaluates metric.
    :param key: The key to retrieve result score from huggingface evaluate result.
    :param generation_gt: A list of ground truth.
        Must be 2-d list of string.
    :param generations: A list of generations that LLM generated.
    :param kwargs: The additional arguments for metric function.
    :return: The list of scores.
    """

    def compute_score(gt: List[str], pred: str) -> float:
        return max(list(map(
            lambda x: instance.compute(predictions=[pred], references=[x], **kwargs)[key], gt)))

    result = list(map(lambda x: compute_score(x[0], x[1]), zip(generation_gt, generations)))
    return result


@convert_inputs_to_list
def bleu(generation_gt: List[List[str]], generations: [str], tokenize: str|None = None, smooth_method: str = 'exp', smooth_value: Optional[float] = None, max_ngram_order: int = 4, trg_lang: str = '', **kwargs) -> List[float]:
    """
    Computes the BLEU metric given pred and ground-truth.

    :param tokenize: The tokenizer to use. If None, defaults to language-specific tokenizers with '13a' as the fallback default. check #https://github.com/mjpost/sacrebleu/blob/master/sacrebleu/metrics/bleu.py
    :param smooth_method: The smoothing method to use ('floor', 'add-k', 'exp' or 'none').
    :param smooth_value: The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
    :param max_ngram_order: If given, it overrides the maximum n-gram order (default: 4) when computing precisions.
    :param trg_lang: An optional language code to raise potential tokenizer warnings.
    :param generation_gt: A list of ground truth.
        Must be 2-d list of string.
        Because it can be a multiple ground truth.
    :param generations: A list of generations that LLM generated.
    """
    bleu = BLEU(tokenize=tokenize, smooth_method=smooth_method, smooth_value=smooth_value, max_ngram_order=max_ngram_order, trg_lang=trg_lang, **kwargs)

    result = list(map(lambda x: bleu.sentence_score(x[0], x[1]).score, zip(generations, generation_gt)))
    return result


@convert_inputs_to_list
def meteor(generation_gt: List[List[str]], generations: List[str],
           alpha: float = 0.9,
           beta: float = 3.0,
           gamma: float = 0.5) -> List[float]:
    """
    Compute meteor score for generation.

    :param generation_gt: A list of ground truth.
            Must be 2-d list of string.
            Because it can be a multiple ground truth.
    :param generations: A list of generations that LLM generated.
    :param alpha: Parameter for controlling relative weights of precision and recall.
        Default is 0.9.
    :param beta: Parameter for controlling shape of penalty as a
        function of as a function of fragmentation.
        Default is 3.0.
    :param gamma: Relative weight assigned to fragmentation penalty.
        Default is 0.5.
    :return: A list of computed metric scores.
    """
    meteor_instance = evaluate.load("meteor")
    result = huggingface_evaluate(meteor_instance, 'meteor', generation_gt, generations,
                                  alpha=alpha, beta=beta, gamma=gamma)
    del meteor_instance
    return result


@convert_inputs_to_list
def rouge(generation_gt: List[List[str]], generations: List[str],
          rouge_type: Optional[str] = 'rougeL',
          use_stemmer: bool = False,
          split_summaries: bool = False,
          batch: int = os.cpu_count()) -> List[float]:
    """
    Compute rouge score for generation.

    :param generation_gt: A list of ground truth.
            Must be 2-d list of string.
            Because it can be a multiple ground truth.
    :param generations: A list of generations that LLM generated.
    :param rouge_type: A rouge type to use for evaluation.
        Default is 'RougeL'.
        Choose between rouge1, rouge2, rougeL, and rougeLSum.
        - rouge1: unigram (1-gram) based scoring.
        - rouge2: bigram (2-gram) based scoring.
        - rougeL: Longest Common Subsequence based scoring.
        - rougeLSum: splits text using "\n"
    :param use_stemmer: Bool indicating whether Porter stemmer should be used to
        strip word suffixes to improve matching. This arg is used in the
        DefaultTokenizer, but other tokenizers might or might not choose to
        use this. Default is False.
    :param split_summaries: Whether to add newlines between sentences for rougeLsum.
        Default is False.
    :param batch: The batch size for processing.
        Default is your cpu count.
    :return: A list of computed metric scores.
    """
    rouge_instance = RougeScorer(rouge_types=[rouge_type], use_stemmer=use_stemmer,
                                 split_summaries=split_summaries,
                                 tokenizer=tokenizers.DefaultTokenizer(use_stemmer))

    async def compute(gt: List[str], pred: str) -> float:
        return rouge_instance.score_multi(targets=gt, prediction=pred)[rouge_type].fmeasure

    tasks = [compute(gt, pred) for gt, pred in zip(generation_gt, generations)]
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(process_batch(tasks, batch_size=batch))

    del rouge_instance
    return result


@convert_inputs_to_list
def sem_score(generation_gt: List[List[str]], generations: List[str],
              embedding_model: Optional[BaseEmbedding] = None,
              batch: int = 128) -> List[float]:
    """
    Compute sem score between generation gt and pred with cosine similarity.

    :param generation_gt: A list of ground truth.
            Must be 2-d list of string.
            Because it can be a multiple ground truth.
            It will get the max of cosine similarity between generation_gt and pred.
    :param generations: A list of generations that LLM generated.
    :param embedding_model: Embedding model to use for compute cosine similarity.
        Default is all-mpnet-base-v2 embedding model.
        The paper used this embedding model.
    :param batch: The batch size for processing.
        Default is 128
    :return: A list of computed metric scores.
    """
    if embedding_model is None:
        embedding_model = embedding_models['huggingface_all_mpnet_base_v2']

    embedding_model.embed_batch_size = batch

    openai_embedding_max_length = 8191
    if isinstance(embedding_model, OpenAIEmbedding):
        generations = openai_truncate_by_token(generations, openai_embedding_max_length,
                                               embedding_model.model_name)

    embedded_pred: List[List[float]] = embedding_model.get_text_embedding_batch(generations,
                                                                                show_progress=True)
    gt_lengths = list(map(len, generation_gt))
    flatten_gt = list(itertools.chain.from_iterable(generation_gt))
    if isinstance(embedding_model, OpenAIEmbedding):
        flatten_gt = openai_truncate_by_token(flatten_gt, openai_embedding_max_length,
                                              embedding_model.model_name)
    embedded_gt_flatten = embedding_model.get_text_embedding_batch(flatten_gt, show_progress=True)
    # re-group embedded_gt_flatten with gt_lengths
    iterator = iter(embedded_gt_flatten)
    embedded_gt: List[List[List[float]]] = [list(itertools.islice(iterator, length)) for length in gt_lengths]

    result = []
    for gt, pred in zip(embedded_gt, embedded_pred):
        similarity_scores: List[float] = list(
            map(lambda x: calculate_cosine_similarity(x, pred), gt))
        result.append(max(similarity_scores))

    del embedding_model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return result


@convert_inputs_to_list
def g_eval(generation_gt: List[List[str]], generations: List[str],
           metrics: Optional[List[str]] = None,
           model: str = 'gpt-4-0125-preview',
           batch_size: int = 8) -> List[float]:
    """
        Calculate G-Eval score.
        G-eval is a metric that uses high-performance LLM model to evaluate generation performance.
        It evaluates the generation result by coherence, consistency, fluency, and relevance.
        It uses only 'openai' model, and we recommend to use gpt-4 for evaluation accuracy.

        :param generation_gt: A list of ground truth.
            Must be 2-d list of string.
            Because it can be a multiple ground truth.
            It will get the max of g_eval score.
        :param generations: A list of generations that LLM generated.
        :param metrics: A list of metrics to use for evaluation.
            Default is all metrics, which is ['coherence', 'consistency', 'fluency', 'relevance'].
        :param model: OpenAI model name.
            Default is 'gpt-4-0125-preview'.
        :param batch_size: The batch size for processing.
            Default is 8.
        :return: G-Eval score.
    """
    loop = asyncio.get_event_loop()
    tasks = [async_g_eval(gt, pred, metrics, model) for gt, pred in zip(generation_gt, generations)]
    result = loop.run_until_complete(process_batch(tasks, batch_size=batch_size))
    return result


async def async_g_eval(generation_gt: List[str], pred: str,
                       metrics: Optional[List[str]] = None,
                       model: str = 'gpt-4-0125-preview',
                       ) -> float:
    available_metrics = ['coherence', 'consistency', 'fluency', 'relevance']
    if metrics is None:
        metrics = available_metrics
    else:
        assert len(metrics) > 0, "metrics must be a list of string"
        metrics = [metric for metric in metrics if metric in available_metrics]

    current_path = os.path.dirname(os.path.realpath(__file__))
    prompt_path = os.path.join(current_path, 'g_eval_prompts')
    g_eval_prompts = {
        "coherence": open(os.path.join(prompt_path, "coh_detailed.txt")).read(),
        "consistency": open(os.path.join(prompt_path, "con_detailed.txt")).read(),
        "fluency": open(os.path.join(prompt_path, "flu_detailed.txt")).read(),
        "relevance": open(os.path.join(prompt_path, "rel_detailed.txt")).read(),
    }

    client = AsyncOpenAI()

    async def g_eval_score(prompt: str, gen_gt: List[str], pred: str):
        scores = []
        for gt in gen_gt:
            input_prompt = prompt.replace('{{Document}}', gt).replace('{{Summary}}', pred)
            response = await client.chat.completions.create(
                model=model,
                messages=[
                    {"role": "system", "content": input_prompt}
                ],
                logprobs=True,
                top_logprobs=5,
                temperature=0,
                max_tokens=2,
                frequency_penalty=0,
                presence_penalty=0,
                stop=None,
                n=20,
            )
            if '(1-3):' in prompt:
                scores.append(get_g_eval_score(response, max_score=3))
            else:
                scores.append(get_g_eval_score(response))
        return max(scores)

    def get_g_eval_score(responses, max_score: int = 5) -> int:
        target_tokens = {str(i): 0 for i in range(1, max_score + 1)}
        for choice in responses.choices:
            first_top_log_probs = choice.logprobs.content[0].top_logprobs
            for i, top_log_prob in enumerate(list(map(lambda x: x.token, first_top_log_probs))):
                if top_log_prob in target_tokens:
                    target_tokens[top_log_prob] += (5 - i)

        return int(max(target_tokens, key=target_tokens.get))

    g_eval_scores = await asyncio.gather(*(g_eval_score(g_eval_prompts[x], generation_gt, pred) for x in metrics))
    return sum(g_eval_scores) / len(g_eval_scores)


@convert_inputs_to_list
def bert_score(generation_gt: List[List[str]], generations: List[str],
               lang: str = 'en',
               batch: int = 128,
               n_threads: int = os.cpu_count()) -> List[float]:
    evaluator = evaluate.load("bertscore")

    df = pd.DataFrame({
        'reference': generation_gt,
        'prediction': generations,
        'lang': lang,
    })

    df = df.explode('reference', ignore_index=False)
    df['bert_score'] = evaluator.compute(predictions=df['prediction'].tolist(),
                                         references=df['reference'].tolist(),
                                         lang=lang, nthreads=n_threads, batch_size=batch)['f1']
    return df.groupby(level=0)['bert_score'].max().tolist()



import pytest

from autorag.nodes.passagereranker import flag_embedding_llm_reranker
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test
from tests.delete_tests import is_github_action


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker():
    top_k = 3
    original_flag_embedding_llm = flag_embedding_llm_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_flag_embedding_llm(queries_example, contents_example, scores_example, ids_example, top_k,
                                      use_fp16=True)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker_batch_one():
    top_k = 3
    batch = 1
    original_flag_embedding_llm = flag_embedding_llm_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_flag_embedding_llm(queries_example, contents_example, scores_example, ids_example, top_k, batch,
                                      use_fp16=True)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker_batch():
    top_k = 3
    batch = 2
    original_flag_embedding_llm = flag_embedding_llm_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_flag_embedding_llm(queries_example, contents_example, scores_example, ids_example, top_k, batch,
                                      use_fp16=True)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker_node():
    top_k = 1
    result_df = flag_embedding_llm_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



import pytest

from autorag.nodes.passagereranker import koreranker
from tests.autorag.nodes.passagereranker.test_passage_reranker_base \
    import (base_reranker_test, base_reranker_node_test, ko_queries_example,
            ko_contents_example, scores_example, ids_example, project_dir, ko_previous_result)
from tests.delete_tests import is_github_action


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_koreranker():
    top_k = 1
    original_koreranker = koreranker.__wrapped__
    contents_result, id_result, score_result \
        = original_koreranker(ko_queries_example, ko_contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k, use_ko=True)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_koreranker_batch_one():
    top_k = 1
    batch = 1
    original_koreranker = koreranker.__wrapped__
    contents_result, id_result, score_result \
        = original_koreranker(ko_queries_example, ko_contents_example, scores_example, ids_example, top_k, batch)
    base_reranker_test(contents_result, id_result, score_result, top_k, use_ko=True)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_koreranker_node():
    top_k = 1
    result_df = koreranker(project_dir=project_dir, previous_result=ko_previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k, use_ko=True)



import pandas as pd

prompt = "Answer this question: {query} \n\n {retrieved_contents}"
queries = ["What is the capital of Japan?", "What is the capital of China?"]
retrieved_contents = [
    ["Tokyo is the capital of Japan.", "Tokyo, the capital of Japan, is a huge metropolitan city."],
    ["Beijing is the capital of China.", "Beijing, the capital of China, is a huge metropolitan city."]]
retrieve_scores = [[0.9, 0.8], [0.9, 0.8]]
retrieved_ids = [["doc1", "doc2"], ["doc3", "doc4"]]
previous_result = pd.DataFrame({
    "query": queries,
    "retrieved_contents": retrieved_contents,
    "retrieve_scores": retrieve_scores,
    "retrieved_ids": retrieved_ids
})

doc_id = ["doc1", "doc2", "doc3", "doc4", "doc5"]
contents = ["This is a test document 1.", "This is a test document 2.", "This is a test document 3.",
            "This is a test document 4.", "This is a test document 5."]
metadata = [{'window': 'havertz arsenal doosan minji naeun gaeun lets go'} for _ in range(5)]
corpus_df = pd.DataFrame({"doc_id": doc_id, "contents": contents, "metadata": metadata})

retrieved_metadata = [
    [{'window': 'havertz arsenal doosan minji naeun gaeun lets go'},
     {'window': 'havertz arsenal doosan minji naeun gaeun lets go'}],
    [{'window': 'havertz arsenal doosan minji naeun gaeun lets go'},
     {'window': 'havertz arsenal doosan minji naeun gaeun lets go'}]
]



from typing import List, Optional

import torch
from llmlingua import PromptCompressor

from autorag.nodes.passagecompressor.base import passage_compressor_node


@passage_compressor_node
def longllmlingua(queries: List[str],
                  contents: List[List[str]],
                  scores,
                  ids,
                  model_name: str = "NousResearch/Llama-2-7b-hf",
                  instructions: Optional[str] = None,
                  target_token: int = 300,
                  **kwargs,
                  ) -> List[str]:
    """
    Compresses the retrieved texts using LongLLMLingua.
    For more information, visit https://github.com/microsoft/LLMLingua.

    :param queries: The queries for retrieved passages.
    :param contents: The contents of retrieved passages.
    :param scores: The scores of retrieved passages.
        Do not use in this function, so you can pass an empty list.
    :param ids: The ids of retrieved passages.
        Do not use in this function, so you can pass an empty list.
    :param model_name: The model name to use for compression.
        Default is "NousResearch/Llama-2-7b-hf".
    :param instructions: The instructions for compression.
        Default is None. When it is None, it will use default instructions.
    :param target_token: The target token for compression.
        Default is 300.
    :param kwargs: Additional keyword arguments.
    :return: The list of compressed texts.
    """
    if instructions is None:
        instructions = "Given the context, please answer the final question"
    llm_lingua = PromptCompressor(
        model_name=model_name,
    )
    results = [llmlingua_pure(query, contents_, llm_lingua, instructions, target_token, **kwargs)
               for query, contents_ in zip(queries, contents)]

    del llm_lingua
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results


def llmlingua_pure(query: str,
                   contents: List[str],
                   llm_lingua: PromptCompressor,
                   instructions: str,
                   target_token: int = 300,
                   **kwargs,
                   ) -> str:
    """
    Return the compressed text.

    :param query: The query for retrieved passages.
    :param contents: The contents of retrieved passages.
    :param llm_lingua: The llm instance that will be used to compress.
    :param instructions: The instructions for compression.
    :param target_token: The target token for compression.
        Default is 300.
    :param kwargs: Additional keyword arguments.
    :return: The compressed text.
    """
    # split by "\n\n" (recommended by LongLLMLingua authors)
    new_context_texts = [c for context in contents for c in context.split("\n\n")]
    compressed_prompt = llm_lingua.compress_prompt(
        new_context_texts,
        question=query,
        instruction=instructions,
        rank_method="longllmlingua",
        target_token=target_token,
        **kwargs,
    )
    compressed_prompt_txt = compressed_prompt["compressed_prompt"]

    # separate out the question and instruction
    result = '\n\n'.join(compressed_prompt_txt.split("\n\n")[1:-1])

    return result



from typing import List, Tuple, Optional

import numpy as np
import torch.cuda

from autorag import embedding_models
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.nodes.passagefilter.base import passage_filter_node
from autorag.utils.util import embedding_query_content


@passage_filter_node
def similarity_threshold_cutoff(queries: List[str], contents_list: List[List[str]],
                                scores_list: List[List[float]], ids_list: List[List[str]],
                                threshold: float, embedding_model: Optional[str] = None,
                                batch: int = 128,
                                ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Re-calculate each content's similarity with the query and filter out the contents that are below the threshold.
    If all contents are filtered, keep the only one highest similarity content.
    This is a filter and does not override scores.
    The output of scores is not coming from query-content similarity.

    :param queries: The list of queries to use for filtering
    :param contents_list: The list of lists of contents to filter
    :param scores_list: The list of lists of scores retrieved
    :param ids_list: The list of lists of ids retrieved
    :param threshold: The threshold to cut off
    :param embedding_model: The embedding model to use for calculating similarity
        Default is OpenAIEmbedding.
    :param batch: The number of queries to be processed in a batch
        Default is 128.
    :return: Tuple of lists containing the filtered contents, ids, and scores
    """
    if embedding_model is None:
        embedding_model = embedding_models['openai']
    else:
        embedding_model = embedding_models[embedding_model]

    query_embeddings, content_embeddings = embedding_query_content(queries, contents_list, embedding_model, batch)

    remain_indices = list(map(lambda x: similarity_threshold_cutoff_pure(x[0], x[1], threshold),
                              zip(query_embeddings, content_embeddings)))

    remain_content_list = list(map(lambda c, idx: [c[i] for i in idx], contents_list, remain_indices))
    remain_scores_list = list(map(lambda s, idx: [s[i] for i in idx], scores_list, remain_indices))
    remain_ids_list = list(map(lambda _id, idx: [_id[i] for i in idx], ids_list, remain_indices))

    del embedding_model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return remain_content_list, remain_ids_list, remain_scores_list


def similarity_threshold_cutoff_pure(query_embedding: str,
                                     content_embeddings: List[List[float]],
                                     threshold: float) -> List[int]:
    """
    Return indices that have to remain.
    Return at least one index if there is nothing to remain.

    :param query_embedding: Query embedding
    :param content_embeddings: Each content embedding
    :param threshold: The threshold to cut off
    :return: Indices to remain at the contents
    """

    similarities = np.array(list(map(lambda x: calculate_cosine_similarity(query_embedding, x),
                                     content_embeddings)))
    result = np.where(similarities >= threshold)[0].tolist()
    if len(result) > 0:
        return result
    return [np.argmax(similarities)]



from .preprocess import (validate_qa_dataset, validate_corpus_dataset, cast_qa_dataset, cast_corpus_dataset,
                         validate_qa_from_corpus_dataset)
from .util import fetch_contents, result_to_dataframe, sort_by_scores



from autorag import embedding_models
from autorag.evaluation.util import cast_metrics


def test_cast_metrics():
    metric1 = ['bleu', 'meteor', 'rouge']
    metric_names, metric_params = cast_metrics(metric1)
    assert metric_names == ['bleu', 'meteor', 'rouge']
    assert metric_params == [{}, {}, {}]

    metric2 = [{'metric_name': 'bleu'}, {'metric_name': 'meteor'}, {'metric_name': 'rouge'}]
    metric_names, metric_params = cast_metrics(metric2)
    assert metric_names == ['bleu', 'meteor', 'rouge']
    assert metric_params == [{}, {}, {}]

    metric3 = [{'metric_name': 'bleu'}, {'metric_name': 'sem_score', 'embedding_model': 'openai'}]
    metric_names, metric_params = cast_metrics(metric3)
    assert metric_names == ['bleu', 'sem_score']
    assert metric_params == [{}, {'embedding_model': embedding_models['openai']}]

    metric4 = [{'metric_name': 'bleu', 'extra_param': 'extra'},
               {'metric_name': 'sem_score', 'embedding_model': 'openai', 'extra_param': 'extra'}]
    metric_names, metric_params = cast_metrics(metric4)
    assert metric_names == ['bleu', 'sem_score']
    assert metric_params == [{'extra_param': 'extra'},
                             {'embedding_model': embedding_models['openai'], 'extra_param': 'extra'}]



import os.path
import pathlib
import tempfile

import pandas as pd
import pytest

from autorag import generator_models
from autorag.nodes.passagecompressor import tree_summarize
from autorag.nodes.passagecompressor.run import run_passage_compressor_node
from autorag.utils.util import load_summary_file
from tests.mock import MockLLM

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resources_dir = os.path.join(root_dir, "resources")
qa_data = pd.DataFrame({
    'qid': ['id-1', 'id-2', 'id-3'],
    'query': ['query-1', 'query-2', 'query-3'],
    'retrieval_gt': [
        [['doc-1', 'doc-2']],
        [['doc-3']],
        [['doc-5'], ['doc-6']],
    ],
    'generation_gt': [['generation-1'], ['generation-2'], ['generation-3']],
})
corpus_data = pd.DataFrame({
    'doc_id': ['doc-1', 'doc-2', 'doc-3', 'doc-4', 'doc-5', 'doc-6'],
    'contents': [
        'Enough for drinking water',
        'Just looking for a water bottle',
        'Do you want to buy some?',
        'I want to buy great octopus',
        'I want to buy some water',
        'Who is son? He is great player in the world',
    ],
    'metadata': [
        {'datetime': '2021-01-01'},
        {'datetime': '2021-01-02'},
        {'datetime': '2021-01-03'},
        {'datetime': '2021-01-04'},
        {'datetime': '2021-01-05'},
        {'datetime': '2021-01-06'},
    ],
})
previous_result = pd.concat([qa_data, pd.DataFrame({
    'retrieved_ids': [
        ['doc-1', 'doc-3'],
        ['doc-2', 'doc-5'],
        ['doc-5', 'doc-6'],
    ],
    'retrieved_contents': [
        ['Enough for drinking water', 'Do you want to buy some?'],
        ['Just looking for a water bottle', 'I want to buy some water'],
        ['I want to buy some water', 'Who is son? He is great player in the world'],
    ],
    'retrieve_scores': [
        [0.1, 0.2],
        [0.3, 0.4],
        [0.5, 0.6],
    ]
})], axis=1)


@pytest.fixture
def node_line_dir():
    with tempfile.TemporaryDirectory() as project_dir:
        data_dir = os.path.join(project_dir, "data")
        os.makedirs(data_dir)
        qa_data.to_parquet(os.path.join(data_dir, "qa.parquet"), index=False)
        corpus_data.to_parquet(os.path.join(data_dir, "corpus.parquet"), index=False)
        trial_dir = os.path.join(project_dir, "trial_1")
        os.makedirs(trial_dir)
        node_line_dir = os.path.join(trial_dir, "node_line_1")
        os.makedirs(node_line_dir)
        yield node_line_dir


def test_run_passage_compressor_node(node_line_dir):
    generator_models['mock'] = MockLLM
    modules = [tree_summarize, tree_summarize]
    module_params = [{'llm': 'mock', 'model': 'gpt-3.5-turbo-16k', 'batch': 4},
                     {'llm': 'mock', 'model': 'gpt-3.5-turbo'}]
    strategies = {
        'metrics': ['retrieval_token_f1', 'retrieval_token_precision'],
        'strategy': 'normalize_mean',
        'speed_threshold': 5,
    }
    best_result = run_passage_compressor_node(modules, module_params, previous_result, node_line_dir, strategies)
    assert os.path.exists(os.path.join(node_line_dir, "passage_compressor"))
    assert set(best_result.columns) == {'qid', 'query', 'retrieval_gt', 'generation_gt',
                                        'retrieved_contents', 'retrieved_ids', 'retrieve_scores',
                                        'passage_compressor_retrieval_token_f1',
                                        'passage_compressor_retrieval_token_precision'}
    # test summary feature
    summary_path = os.path.join(node_line_dir, "passage_compressor", "summary.csv")
    single_result_path = os.path.join(node_line_dir, "passage_compressor", '0.parquet')
    assert os.path.exists(single_result_path)
    single_result_df = pd.read_parquet(single_result_path)
    assert os.path.exists(summary_path)
    summary_df = load_summary_file(summary_path)
    assert set(summary_df.columns) == {'filename', 'passage_compressor_retrieval_token_f1',
                                       'passage_compressor_retrieval_token_precision',
                                       'module_name', 'module_params', 'execution_time', 'is_best'}
    assert len(summary_df) == 2
    assert summary_df['filename'][0] == "0.parquet"
    assert summary_df['passage_compressor_retrieval_token_f1'][0] == pytest.approx(
        single_result_df['retrieval_token_f1'].mean())
    assert summary_df['passage_compressor_retrieval_token_precision'][0] == pytest.approx(
        single_result_df['retrieval_token_precision'].mean())
    assert summary_df['module_name'][0] == "tree_summarize"
    assert summary_df['module_params'][0] == {'llm': 'mock', 'model': 'gpt-3.5-turbo-16k', 'batch': 4}
    assert summary_df['execution_time'][0] > 0
    # test the best file is saved properly
    best_path = summary_df[summary_df['is_best']]['filename'].values[0]
    assert os.path.exists(os.path.join(node_line_dir, "passage_compressor", f'best_{best_path}'))



"""
This file contains the retrieval contents metric,
which means calculate the metric based on the contents of the retrieved items.
"""
import functools
import itertools
from collections import Counter
from typing import List

import numpy as np

from autorag.utils.util import normalize_string, convert_inputs_to_list


def retrieval_contents_metric(func):
    @functools.wraps(func)
    @convert_inputs_to_list
    def wrapper(gt_contents: List[List[str]], pred_contents: List[List[str]]) -> List[float]:
        results = []
        for gt, pred in zip(gt_contents, pred_contents):
            if gt == [] or any(bool(g) is False for g in gt):
                results.append(None)
            else:
                results.append(func(gt, pred))
        return results

    return wrapper


def single_token_f1(ground_truth: str, prediction: str):
    prediction_tokens = normalize_string(prediction).split()
    ground_truth_tokens = normalize_string(ground_truth).split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0, 0, 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return precision, recall, f1


@retrieval_contents_metric
def retrieval_token_f1(gt: List[str], pred: List[str]):
    calculated_results = list(map(lambda x: single_token_f1(x[1], x[0]), list(itertools.product(pred, gt))))
    _, _, result = zip(*calculated_results)
    result_np = np.array(list(result)).reshape(len(pred), -1)
    return result_np.max(axis=1).mean()


@retrieval_contents_metric
def retrieval_token_precision(gt: List[str], pred: List[str]):
    calculated_results = list(map(lambda x: single_token_f1(x[1], x[0]), list(itertools.product(pred, gt))))
    result, _, _ = zip(*calculated_results)
    result_np = np.array(list(result)).reshape(len(pred), -1)
    return result_np.max(axis=1).mean()


@retrieval_contents_metric
def retrieval_token_recall(gt: List[str], pred: List[str]):
    calculated_results = list(map(lambda x: single_token_f1(x[1], x[0]), list(itertools.product(pred, gt))))
    _, result, _ = zip(*calculated_results)
    result_np = np.array(list(result)).reshape(len(pred), -1)
    return result_np.max(axis=1).mean()



import os
import tempfile
from datetime import datetime

import chromadb
import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.retrieval.bm25 import bm25_ingest
from autorag.nodes.retrieval.vectordb import vectordb_ingest

sample_ids = ([
                  ['id-1', 'id-2', 'id-3'],
                  ['id-2', 'id-3', 'id-4']
              ], [
                  ['id-1', 'id-4', 'id-3'],
                  ['id-2', 'id-5', 'id-4']
              ])
sample_scores = ([
                     [1, 3, 5],
                     [2, 4, 6]
                 ], [
                     [4, 2, 6],
                     [5, 3, 7]
                 ])
sample_ids_2 = (['id-1', 'id-2', 'id-3', 'id-4', 'id-5'],
                ['id-1', 'id-4', 'id-3', 'id-5', 'id-2'])
sample_scores_2 = ([5, 3, 1, 0.4, 0.2], [6, 2, 1, 0.5, 0.1])

sample_ids_3 = ([
                    ['id-1', 'id-2', 'id-3', 'id-4', 'id-5'],
                    ['id-2', 'id-3', 'id-4', 'id-5', 'id-6']
                ], [
                    ['id-1', 'id-4', 'id-3', 'id-5', 'id-2'],
                    ['id-2', 'id-5', 'id-4', 'id-6', 'id-3']
                ])
sample_scores_3 = ([
                       [5, 3, 1, 0.4, 0.2],
                       [6, 4, 2, 1.4, 1.2]
                   ], [
                       [6, 2, 1, 0.5, 0.1],
                       [7, 3, 2, 1.5, 1.1],
                   ])

sample_ids_non_overlap = (['id-1', 'id-2', 'id-3', 'id-4', 'id-5'],
                          ['id-6', 'id-4', 'id-3', 'id-7', 'id-2'])

previous_result = pd.DataFrame({
    'qid': ['query-1', 'query-2', 'query-3'],
    'query': ['query-1', 'query-2', 'query-3'],
    'retrieval_gt': [
        [['id-1'], ['id-2'], ['id-3']],
        [['id-1'], ['id-2'], ['id-3']],
        [['id-1'], ['id-2'], ['id-3']]
    ],
    'generation_gt': [
        ['gen-1', 'gen-2'],
        ['gen-1', 'gen-2'],
        ['gen-1', 'gen-2']
    ]
})

modules_with_weights = {
    'ids': ([['id-1', 'id-2', 'id-3', 'id-4', 'id-5'],
             ['id-1', 'id-2', 'id-3', 'id-4', 'id-5']],
            [['id-1', 'id-4', 'id-3', 'id-5', 'id-2'],
             ['id-1', 'id-4', 'id-3', 'id-5', 'id-2']]
            ),
    'scores': ([[5, 3, 1, 0.4, 0.2], [5, 3, 1, 0.4, 0.2]],
               [[6, 2, 1, 0.5, 0.1], [6, 2, 1, 0.5, 0.1]]),
    'top_k': 3,
    'weights': (0.3, 0.7)
}


@pytest.fixture
def pseudo_project_dir():
    with tempfile.TemporaryDirectory() as project_dir:
        corpus_df = pd.DataFrame({
            'doc_id': ['id-1', 'id-2', 'id-3', 'id-4', 'id-5', 'id-6', 'id-7', 'id-8', 'id-9'],
            'contents': ['doc-1', 'doc-2', 'doc-3', 'doc-4', 'doc-5', 'doc-6', 'doc-7', 'doc-8', 'doc-9'],
            'metadata': [{'last_modified_date': datetime.now()} for _ in range(9)]
        })
        os.makedirs(os.path.join(project_dir, "data"))
        corpus_df.to_parquet(os.path.join(project_dir, "data", 'corpus.parquet'))
        resource_dir = os.path.join(project_dir, "resources")
        os.makedirs(resource_dir)
        bm25_ingest(os.path.join(resource_dir, 'bm25_porter_stemmer.pkl'), corpus_df)
        chroma_path = os.path.join(resource_dir, 'chroma')
        db = chromadb.PersistentClient(path=chroma_path)
        collection = db.create_collection(name="openai", metadata={"hnsw:space": "cosine"})
        vectordb_ingest(collection, corpus_df, OpenAIEmbedding())
        yield project_dir


def base_hybrid_weights_node_test(hybrid_func, pseudo_project_dir, retrieve_scores):
    result = hybrid_func(project_dir=pseudo_project_dir, previous_result=previous_result, **modules_with_weights)
    assert len(result) == 2
    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {'retrieved_contents', 'retrieved_ids', 'retrieve_scores'}
    assert result['retrieved_ids'].tolist()[0] == ['id-1', 'id-4', 'id-2']
    assert result['retrieve_scores'].tolist()[0] == pytest.approx(retrieve_scores)
    assert result['retrieved_contents'].tolist()[0] == ['doc-1', 'doc-4', 'doc-2']



from autorag.nodes.passagefilter import similarity_percentile_cutoff
from tests.autorag.nodes.passagefilter.test_passage_filter_base import queries_example, contents_example, \
    scores_example, ids_example, base_passage_filter_test, project_dir, previous_result, base_passage_filter_node_test


def test_similarity_percentile_cutoff():
    original_cutoff = similarity_percentile_cutoff.__wrapped__
    contents, ids, scores = original_cutoff(
        queries_example, contents_example, scores_example, ids_example, percentile=0.85,
        embedding_model='openai_embed_3_large', batch=64)
    num_top_k = int(len(contents_example[0]) * 0.85)
    assert len(contents[0]) == len(contents[1]) == num_top_k
    base_passage_filter_test(contents, ids, scores)


def test_similarity_percentile_cutoff_node():
    result_df = similarity_percentile_cutoff(
        project_dir=project_dir, previous_result=previous_result, percentile=0.9)
    base_passage_filter_node_test(result_df)



import os
import pathlib
import shutil
import tempfile
import uuid
from datetime import datetime
from unittest.mock import patch

import chromadb
import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding, OpenAIEmbeddingModelType

from autorag.nodes.retrieval import vectordb
from autorag.nodes.retrieval.vectordb import vectordb_ingest
from tests.autorag.nodes.retrieval.test_retrieval_base import (queries, corpus_df, previous_result,
                                                               base_retrieval_test, base_retrieval_node_test)

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resource_path = os.path.join(root_dir, "resources")

embedding_model = OpenAIEmbedding(model_name=OpenAIEmbeddingModelType.TEXT_EMBED_3_SMALL)


@pytest.fixture
def ingested_vectordb():
    with tempfile.TemporaryDirectory() as chroma_path:
        db = chromadb.PersistentClient(path=chroma_path)
        collection = db.create_collection(name="test_vectordb_retrieval", metadata={"hnsw:space": "cosine"})

        vectordb_ingest(collection, corpus_df, embedding_model)

        assert collection.count() == 5
        yield collection


@pytest.fixture
def empty_chromadb():
    with tempfile.TemporaryDirectory() as chroma_path:
        db = chromadb.PersistentClient(path=chroma_path)
        collection = db.create_collection(name="test_vectordb_retrieval", metadata={"hnsw:space": "cosine"})

        yield collection


@pytest.fixture
def project_dir_for_vectordb_node():
    with tempfile.TemporaryDirectory() as test_project_dir:
        sample_project_dir = os.path.join(resource_path, "sample_project")
        # copy & paste all folders and files in sample_project folder
        shutil.copytree(sample_project_dir, test_project_dir, dirs_exist_ok=True)

        chroma_path = os.path.join(test_project_dir, "resources", "chroma")
        os.makedirs(chroma_path)
        db = chromadb.PersistentClient(path=chroma_path)
        collection = db.create_collection(name="openai", metadata={"hnsw:space": "cosine"})
        corpus_path = os.path.join(test_project_dir, "data", "corpus.parquet")
        corpus_df = pd.read_parquet(corpus_path)
        vectordb_ingest(collection, corpus_df, embedding_model)

        yield test_project_dir


def test_vectordb_retrieval(ingested_vectordb):
    top_k = 4
    original_vectordb = vectordb.__wrapped__
    id_result, score_result = original_vectordb(queries, top_k=top_k, collection=ingested_vectordb,
                                                embedding_model=embedding_model)
    base_retrieval_test(id_result, score_result, top_k)


def test_vectordb_node(project_dir_for_vectordb_node):
    result_df = vectordb(project_dir=project_dir_for_vectordb_node, previous_result=previous_result, top_k=4,
                         embedding_model="openai")
    base_retrieval_node_test(result_df)


def test_duplicate_id_vectordb_ingest(ingested_vectordb):
    vectordb_ingest(ingested_vectordb, corpus_df, embedding_model)
    assert ingested_vectordb.count() == 5

    new_doc_id = ["doc4", "doc5", "doc6", "doc7", "doc8"]
    new_contents = ["This is a test document 4.", "This is a test document 5.", "This is a test document 6.",
                    "This is a test document 7.", "This is a test document 8."]
    new_metadata = [{'datetime': datetime.now()} for _ in range(5)]
    new_corpus_df = pd.DataFrame({"doc_id": new_doc_id, "contents": new_contents, "metadata": new_metadata})
    vectordb_ingest(ingested_vectordb, new_corpus_df, embedding_model)

    assert ingested_vectordb.count() == 8


def test_long_text_vectordb_ingest(ingested_vectordb):
    new_doc_id = ["doc6", "doc7"]
    new_contents = ["This is a test" * 20000,
                    "This is a test" * 40000]
    new_metadata = [{'datetime': datetime.now()} for _ in range(2)]
    new_corpus_df = pd.DataFrame({"doc_id": new_doc_id, "contents": new_contents, "metadata": new_metadata})
    assert isinstance(embedding_model, OpenAIEmbedding)
    vectordb_ingest(ingested_vectordb, new_corpus_df, embedding_model)

    assert ingested_vectordb.count() == 7


def mock_get_text_embedding_batch(self, texts, **kwargs):
    return [[3.0, 4.1, 3.2] for _ in range(len(texts))]


@patch.object(OpenAIEmbedding, 'get_text_embedding_batch', mock_get_text_embedding_batch)
def test_long_ids_ingest(empty_chromadb):
    embedding_model = OpenAIEmbedding()
    content_df = pd.DataFrame({
        'doc_id': [str(uuid.uuid4()) for _ in range(10000)],
        'contents': ['havertz' for _ in range(10000)],
        'metadata': [{'last_modified_datetime': datetime.now()} for _ in range(10000)],
    })
    vectordb_ingest(empty_chromadb, content_df, embedding_model)



from autorag.nodes.passagefilter import percentile_cutoff
from tests.autorag.nodes.passagefilter.test_passage_filter_base import queries_example, contents_example, \
    scores_example, ids_example, base_passage_filter_test, project_dir, previous_result, base_passage_filter_node_test


def test_percentile_cutoff():
    original_cutoff = percentile_cutoff.__wrapped__
    contents, ids, scores = original_cutoff(
        queries_example, contents_example, scores_example, ids_example, percentile=0.6)
    base_passage_filter_test(contents, ids, scores)
    assert scores[0] == [0.8, 0.5]
    assert contents[0] == ["Paris is the capital of France.",
                           "Paris is one of the capital from France. Isn't it?"]


def test_percentile_cutoff_reverse():
    original_cutoff = percentile_cutoff.__wrapped__
    contents, ids, scores = original_cutoff(
        queries_example, contents_example, scores_example, ids_example, percentile=0.6, reverse=True)
    base_passage_filter_test(contents, ids, scores)
    assert scores[0] == [0.1, 0.1]
    assert contents[0] == ["NomaDamas is Great Team", "havertz is suck at soccer"]


def test_percentile_cutoff_node():
    result_df = percentile_cutoff(
        project_dir=project_dir, previous_result=previous_result, percentile=0.9)
    base_passage_filter_node_test(result_df)



import asyncio
import logging
import os
from typing import List, Tuple

import tiktoken
from openai import AsyncOpenAI
from tiktoken import Encoding

from autorag.nodes.generator.base import generator_node
from autorag.utils.util import process_batch

logger = logging.getLogger("AutoRAG")

MAX_TOKEN_DICT = {  # model name : token limit
    'gpt-4o': 128_000,
    'gpt-4o-2024-05-13': 128_000,
    'gpt-4-turbo': 128_000,
    'gpt-4-turbo-2024-04-09': 128_000,
    'gpt-4-turbo-preview': 128_000,
    'gpt-4-0125-preview': 128_000,
    'gpt-4-1106-preview': 128_000,
    'gpt-4-vision-preview': 128_000,
    'gpt-4-1106-vision-preview': 128_000,
    'gpt-4': 8_192,
    'gpt-4-0613': 8_192,
    'gpt-4-32k': 32_768,
    'gpt-4-32k-0613': 32_768,
    'gpt-3.5-turbo-0125': 16_385,
    'gpt-3.5-turbo': 16_385,
    'gpt-3.5-turbo-1106': 16_385,
    'gpt-3.5-turbo-instruct': 4_096,
    'gpt-3.5-turbo-16k': 16_385,
    'gpt-3.5-turbo-0613': 4_096,
    'gpt-3.5-turbo-16k-0613': 16_385,
}


@generator_node
def openai_llm(prompts: List[str], llm: str = "gpt-3.5-turbo", batch: int = 16,
               truncate: bool = True,
               api_key: str = None,
               **kwargs) -> \
        Tuple[List[str], List[List[int]], List[List[float]]]:
    """
    OpenAI generator module.
    Uses official openai library for generating answer from the given prompt.
    It returns real token ids and log probs, so you must use this for using token ids and log probs.

    :param prompts: A list of prompts.
    :param llm: A model name for openai.
        Default is gpt-3.5-turbo.
    :param batch: Batch size for openai api call.
        If you get API limit errors, you should lower the batch size.
        Default is 16.
    :param truncate: Whether to truncate the input prompt.
        Default is True.
    :param api_key: OpenAI API key. You can set this by passing env variable `OPENAI_API_KEY`
    :param kwargs: The optional parameter for openai api call `openai.chat.completion`
        See https://platform.openai.com/docs/api-reference/chat/create for more details.
    :return: A tuple of three elements.
        The first element is a list of generated text.
        The second element is a list of generated text's token ids.
        The third element is a list of generated text's log probs.
    """
    if api_key is None:
        api_key = os.getenv("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError("OPENAI_API_KEY does not set. "
                             "Please set env variable OPENAI_API_KEY or pass api_key parameter to openai module.")

    if kwargs.get('logprobs') is not None:
        kwargs.pop('logprobs')
        logger.warning("parameter logprob does not effective. It always set to True.")
    if kwargs.get('n') is not None:
        kwargs.pop('n')
        logger.warning("parameter n does not effective. It always set to 1.")

    tokenizer = tiktoken.encoding_for_model(llm)
    if truncate:
        max_token_size = MAX_TOKEN_DICT.get(llm) - 7  # because of chat token usage
        if max_token_size is None:
            raise ValueError(f"Model {llm} does not supported. "
                             f"Please select the model between {list(MAX_TOKEN_DICT.keys())}")
        prompts = list(map(lambda prompt: truncate_by_token(prompt, tokenizer, max_token_size), prompts))

    client = AsyncOpenAI(api_key=api_key)
    loop = asyncio.get_event_loop()
    tasks = [get_result(prompt, client, llm, tokenizer, **kwargs) for prompt in prompts]
    result = loop.run_until_complete(process_batch(tasks, batch))
    answer_result = list(map(lambda x: x[0], result))
    token_result = list(map(lambda x: x[1], result))
    logprob_result = list(map(lambda x: x[2], result))
    return answer_result, token_result, logprob_result


async def get_result(prompt: str, client: AsyncOpenAI, model: str, tokenizer: Encoding, **kwargs):
    response = await client.chat.completions.create(
        model=model,
        messages=[
            {"role": "user", "content": prompt},
        ],
        logprobs=True,
        n=1,
        **kwargs
    )
    choice = response.choices[0]
    answer = choice.message.content
    logprobs = list(map(lambda x: x.logprob, choice.logprobs.content))
    tokens = list(map(lambda x: tokenizer.encode(x.token, allowed_special='all')[0], choice.logprobs.content))
    assert len(tokens) == len(logprobs), "tokens and logprobs size is different."
    return answer, tokens, logprobs


def truncate_by_token(prompt: str, tokenizer: Encoding, max_token_size: int):
    tokens = tokenizer.encode(prompt, allowed_special='all')
    return tokenizer.decode(tokens[:max_token_size])



from typing import Tuple, List

import pandas as pd
import pytest

from autorag.evaluation import evaluate_retrieval_contents

gt = [
    ['Enough for drinking water', 'Just looking for a water bottle'],
    ['Do you want to buy some?']
]
pred = [
    ['Enough for mixing water', 'I want to do a nothing', 'Just looking is a very healthy'],
    ['Do you want to buy some?', 'I want to buy some', 'I want to buy some water']
]


@evaluate_retrieval_contents(retrieval_gt=gt,
                             metrics=['retrieval_token_recall', 'retrieval_token_precision', 'retrieval_token_f1'])
def pseudo_retrieval() -> Tuple[List[List[str]], List[List[float]], List[List[str]]]:
    return pred, [[0.3, 0.2, 0.1], [0.3, 0.2, 0.1]], [['pred-0', 'pred-1', 'pred-2'], ['pred-3', 'pred-4', 'pred-5']]


def test_evaluate_retrieval_contents():
    result_df = pseudo_retrieval()
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == 2
    assert set(result_df.columns) == {
        'retrieved_contents', 'retrieved_ids', 'retrieve_scores',
        'retrieval_token_recall', 'retrieval_token_precision', 'retrieval_token_f1'
    }
    assert result_df['retrieval_token_recall'].tolist() == pytest.approx([0.383333, 0.777777], rel=0.001)
    assert result_df['retrieval_token_precision'].tolist() == pytest.approx([0.383333, 0.8222222], rel=0.001)
    assert result_df['retrieval_token_f1'].tolist() == pytest.approx([0.38333, 0.797979], rel=0.001)




import asyncio
from typing import List, Optional, Sequence, Tuple

import numpy as np
from llama_index.core.base.llms.types import ChatMessage, ChatResponse
from llama_index.core.llms import LLM
from llama_index.core.postprocessor.rankGPT_rerank import RankGPTRerank
from llama_index.core.schema import NodeWithScore, QueryBundle, TextNode
from llama_index.core.utils import print_text
from llama_index.llms.openai import OpenAI

from autorag import generator_models
from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import process_batch


@passage_reranker_node
def rankgpt(queries: List[str], contents_list: List[List[str]],
            scores_list: List[List[float]], ids_list: List[List[str]],
            top_k: int, llm: Optional[LLM] = None, verbose: bool = False,
            rankgpt_rerank_prompt: Optional[str] = None,
            batch: int = 16, **kwargs) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank given context paragraphs using RankGPT.
    Return pseudo scores, since the actual scores are not available on RankGPT.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param llm: The LLM model to use for RankGPT rerank.
        It is LlamaIndex model.
        Default is OpenAI model with gpt-3.5-turbo-16k.
    :param verbose: Whether to print intermediate steps.
    :param rankgpt_rerank_prompt: The prompt template for RankGPT rerank.
        Default is RankGPT's default prompt.
    :param batch: The number of queries to be processed in a batch.
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    query_bundles = list(map(lambda query: QueryBundle(query_str=query), queries))
    nodes_list = [
        list(map(lambda x: NodeWithScore(node=TextNode(text=x[0]), score=x[1]), zip(content_list, score_list)))
        for content_list, score_list in zip(contents_list, scores_list)]
    if llm is None:
        llm = OpenAI(model="gpt-3.5-turbo-16k")

    if not isinstance(llm, LLM):
        llm = generator_models[llm](**kwargs)

    reranker = AsyncRankGPTRerank(top_n=top_k, llm=llm, verbose=verbose, rankgpt_rerank_prompt=rankgpt_rerank_prompt)

    tasks = [reranker.async_postprocess_nodes(nodes, query, ids) for nodes, query, ids in
             zip(nodes_list, query_bundles, ids_list)]
    loop = asyncio.get_event_loop()
    rerank_result = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    content_result = [list(map(lambda x: x.node.text, res[0])) for res in rerank_result]
    score_result = [np.linspace(1.0, 0.0, len(res[0])).tolist() for res in rerank_result]
    id_result = [res[1] for res in rerank_result]

    del reranker
    del llm

    return content_result, id_result, score_result


class AsyncRankGPTRerank(RankGPTRerank):
    async def async_run_llm(self, messages: Sequence[ChatMessage]) -> ChatResponse:
        return await self.llm.achat(messages)

    async def async_postprocess_nodes(self,
                                      nodes: List[NodeWithScore],
                                      query_bundle: QueryBundle,
                                      ids: Optional[List[str]] = None,
                                      ) -> Tuple[List[NodeWithScore], List[str]]:
        if ids is None:
            ids = [str(i) for i in range(len(nodes))]

        items = {
            "query": query_bundle.query_str,
            "hits": [{"content": node.get_content()} for node in nodes],
        }

        messages = self.create_permutation_instruction(item=items)
        permutation = await self.async_run_llm(messages=messages)
        if permutation.message is not None and permutation.message.content is not None:
            rerank_ranks = self._receive_permutation(
                items, str(permutation.message.content)
            )
            if self.verbose:
                print_text(f"After Reranking, new rank list for nodes: {rerank_ranks}")

            initial_results: List[NodeWithScore] = []
            id_results = []

            for idx in rerank_ranks:
                initial_results.append(
                    NodeWithScore(node=nodes[idx].node, score=nodes[idx].score)
                )
                id_results.append(ids[idx])
            return initial_results[: self.top_n], id_results[: self.top_n]
        else:
            return nodes[: self.top_n], ids[: self.top_n]



import uuid
from typing import Optional

import pandas as pd
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from ragas.testset import TestsetGenerator
from ragas.testset.evolutions import simple, reasoning, multi_context

from autorag.data.utils.util import corpus_df_to_langchain_documents
from autorag.utils import cast_qa_dataset


def generate_qa_ragas(corpus_df: pd.DataFrame,
                      test_size: int,
                      distributions: Optional[dict] = None,
                      generator_llm: Optional[BaseChatModel] = None,
                      critic_llm: Optional[BaseChatModel] = None,
                      embedding_model: Optional[Embeddings] = None,
                      **kwargs
                      ) -> pd.DataFrame:
    """
    QA dataset generation using RAGAS.
    Returns qa dataset dataframe.

    :param corpus_df: Corpus dataframe.
    :param test_size: Number of queries to generate.
    :param distributions: Distributions of different types of questions.
        Default is "simple is 0.5, multi_context is 0.4, and reasoning is 0.1."
        Each type of questions refers to Ragas evolution types.
    :param generator_llm: Generator language model from Langchain.
    :param critic_llm: Critic language model from Langchain.
    :param embedding_model: Embedding model from Langchain.
    :param kwargs: The additional option to pass to the 'generate_with_langchain_docs' method.
        You can input 'with_debugging_logs', 'is_async', 'raise_exceptions', and 'run_config'.
    :return: QA dataset dataframe.
    """
    if generator_llm is None:
        generator_llm = ChatOpenAI(model="gpt-3.5-turbo-16k")
    if critic_llm is None:
        critic_llm = ChatOpenAI(model="gpt-4-turbo")
    if embedding_model is None:
        embedding_model = OpenAIEmbeddings()
    if distributions is None:
        distributions = {
            simple: 0.5,
            multi_context: 0.4,
            reasoning: 0.1
        }

    assert sum(list(distributions.values())) == 1.0, "Sum of distributions must be 1.0"

    generator = TestsetGenerator.from_langchain(generator_llm, critic_llm, embedding_model)

    langchain_docs = corpus_df_to_langchain_documents(corpus_df)

    test_df = generator.generate_with_langchain_docs(langchain_docs, test_size, distributions=distributions,
                                                     **kwargs).to_pandas()

    result_df = pd.DataFrame({
        'qid': [str(uuid.uuid4()) for _ in range(len(test_df))],
        'query': test_df['question'].tolist(),
        'generation_gt': list(map(lambda x: x, test_df['ground_truth'].tolist())),
    })

    result_df['retrieval_gt'] = test_df['metadata'].apply(lambda x: list(map(lambda y: y['filename'], x)))
    result_df = cast_qa_dataset(result_df)

    return result_df



from typing import Optional

import click
import streamlit as st

from autorag.deploy import Runner


def get_runner(yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]):
    if not yaml_path and not trial_path:
        raise ValueError("yaml_path or trial_path must be given.")
    elif yaml_path and trial_path:
        raise ValueError("yaml_path and trial_path cannot be given at the same time.")
    elif yaml_path:
        return Runner.from_yaml(yaml_path, project_dir=project_dir)
    elif trial_path:
        return Runner.from_trial_folder(trial_path)


def set_initial_state():
    if "messages" not in st.session_state:
        st.session_state["messages"] = [
            {
                "role": "assistant",
                "content": "Welcome !",
            }
        ]


def set_page_config():
    st.set_page_config(
        page_title="AutoRAG",
        page_icon="🤖",
        layout="wide",
        initial_sidebar_state="expanded",
        menu_items={
            'Get help': "https://github.com/Marker-Inc-Korea/AutoRAG/discussions",
            'Report a bug': "https://github.com/Marker-Inc-Korea/AutoRAG/issues",
        }
    )


def set_page_header():
    st.header("📚 AutoRAG", anchor=False)
    st.caption(
        "Input a question and get an answer from the given documents. "
    )


def chat_box(runner: Runner):
    if query := st.chat_input("How can I help?"):
        # Add the user input to messages state
        st.session_state["messages"].append({"role": "user", "content": query})
        with st.chat_message("user"):
            st.markdown(query)

        # Generate llama-index stream with user input
        with st.chat_message("assistant"):
            with st.spinner("Processing..."):
                response = st.write(runner.run(query))

        # Add the final response to messages state
        st.session_state["messages"].append({"role": "assistant", "content": response})


@click.command()
@click.option("--yaml_path", type=str, help="Path to the YAML file.")
@click.option("--project_dir", type=str, help="Path to the project directory.")
@click.option("--trial_path", type=str, help="Path to the trial directory.")
def run_web_server(yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]):
    import nest_asyncio
    nest_asyncio.apply()
    runner = get_runner(yaml_path, project_dir, trial_path)
    set_initial_state()
    set_page_config()
    set_page_header()
    chat_box(runner)


if __name__ == "__main__":
    run_web_server()



import functools
import logging
import os
from pathlib import Path
from typing import List, Union, Tuple

import numpy as np
import pandas as pd

from autorag import embedding_models
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.utils import (result_to_dataframe, validate_qa_dataset, fetch_contents, sort_by_scores,
                           validate_corpus_dataset, cast_corpus_dataset)
from autorag.utils.util import filter_dict_keys, select_top_k, embedding_query_content

logger = logging.getLogger("AutoRAG")


def passage_augmenter_node(func):
    @functools.wraps(func)
    @result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
        logger.info(f"Running passage augmenter node - {func.__name__} module...")
        validate_qa_dataset(previous_result)
        data_dir = os.path.join(project_dir, "data")

        # find queries columns
        assert "query" in previous_result.columns, "previous_result must have query column."
        queries = previous_result["query"].tolist()

        # find ids columns
        assert "retrieved_ids" in previous_result.columns, "previous_result must have retrieved_ids column."
        ids = previous_result["retrieved_ids"].tolist()

        corpus_df = pd.read_parquet(os.path.join(data_dir, "corpus.parquet"), engine='pyarrow')
        validate_corpus_dataset(corpus_df)

        # get top_k
        top_k = kwargs.pop("top_k")

        if func.__name__ == 'prev_next_augmenter':
            corpus_df = cast_corpus_dataset(corpus_df)
            slim_corpus_df = corpus_df[["doc_id", "metadata"]]
            slim_corpus_df['metadata'] = slim_corpus_df['metadata'].apply(filter_dict_keys, keys=['prev_id', 'next_id'])

            mode = kwargs.pop("mode", 'both')
            num_passages = kwargs.pop("num_passages", 1)

            # get augmented ids
            augmented_ids = func(ids_list=ids, corpus_df=slim_corpus_df, mode=mode, num_passages=num_passages)

            # fetch contents from corpus to use augmented ids
            augmented_contents = fetch_contents(corpus_df, augmented_ids)

            # set embedding model for getting scores
            embedding_model_str = kwargs.pop("embedding_model", 'openai')
            embedding_model = embedding_models[embedding_model_str]
            query_embeddings, contents_embeddings = embedding_query_content(queries, augmented_contents,
                                                                            embedding_model, batch=128)

            # get scores from calculated cosine similarity
            augmented_scores = [
                np.array([calculate_cosine_similarity(query_embedding, x) for x in content_embeddings]).tolist()
                for query_embedding, content_embeddings in zip(query_embeddings, contents_embeddings)]
        else:
            contents = previous_result["retrieved_contents"].tolist()
            scores = previous_result["retrieve_scores"].tolist()

            augmented_ids, augmented_contents, augmented_scores = func(ids_list=ids, contents_list=contents,
                                                                       scores_list=scores, **kwargs)

        # sort by scores
        df = pd.DataFrame({
            'contents': augmented_contents,
            'ids': augmented_ids,
            'scores': augmented_scores,
        })
        df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')

        # select by top_k
        results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

        return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()

    return wrapper



import os.path
import pathlib
import shutil
import tempfile
from unittest.mock import patch

import pandas as pd
import pytest
from llama_index.core.base.llms.types import CompletionResponse
from llama_index.llms.openai import OpenAI

from autorag.nodes.queryexpansion import query_decompose, hyde
from autorag.nodes.queryexpansion.run import evaluate_one_query_expansion_node
from autorag.nodes.queryexpansion.run import run_query_expansion_node
from autorag.nodes.retrieval import bm25
from autorag.utils.util import load_summary_file

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resources_dir = os.path.join(root_dir, "resources")

sample_expanded_queries = [
    ["What is the purpose of rushing up the middle in football?",
     "Why are the first two plays often used for rushing up the middle in football?",
     "What are regular rush plays in football and how do they differ from rushing up the middle?"
     ],
    ["What are the prices of regular, mid, and premium gas?",
     " Why is there a 10 cent difference between the prices of regular, mid, and premium gas?",
     " Are there any specific factors or regulations that determine the pricing tiers of gas?"]
]
metrics = ['retrieval_f1', 'retrieval_recall']


@pytest.fixture
def node_line_dir():
    with tempfile.TemporaryDirectory() as project_dir:
        sample_project_dir = os.path.join(resources_dir, "sample_project")
        # copy & paste all folders and files in sample_project folder
        shutil.copytree(sample_project_dir, project_dir, dirs_exist_ok=True)

        test_trail_dir = os.path.join(project_dir, "test_trial")
        os.makedirs(test_trail_dir)
        node_line_dir = os.path.join(test_trail_dir, "test_node_line")
        os.makedirs(node_line_dir)
        yield node_line_dir


def test_evaluate_one_query_expansion_node(node_line_dir):
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    previous_result = pd.read_parquet(qa_path)
    sample_previous_result = previous_result.head(2)
    sample_retrieval_gt = sample_previous_result['retrieval_gt'].tolist()

    retrieval_funcs = [bm25, bm25]
    retrieval_params = [{'top_k': 1, 'bm25_tokenizer': 'gpt2'}, {'top_k': 2, 'bm25_tokenizer': 'gpt2'}]
    best_result = evaluate_one_query_expansion_node(retrieval_funcs, retrieval_params,
                                                    sample_expanded_queries, sample_retrieval_gt,
                                                    metrics, project_dir, sample_previous_result, 'rank')
    assert isinstance(best_result, pd.DataFrame)
    assert all(metric_name in best_result.columns for metric_name in metrics)
    assert len(best_result) == len(sample_expanded_queries)


def base_query_expansion_test(best_result, node_line_dir):
    assert os.path.exists(os.path.join(node_line_dir, "query_expansion"))
    expect_columns = ['qid', 'query', 'generation_gt', 'retrieval_gt', 'queries',
                      'query_expansion_retrieval_f1', 'query_expansion_retrieval_recall']
    assert all([expect_column in best_result.columns for expect_column in expect_columns])
    assert os.path.exists(os.path.join(node_line_dir, "query_expansion", "0.parquet"))
    assert os.path.exists(os.path.join(node_line_dir, "query_expansion", "1.parquet"))
    # test summary feature
    summary_path = os.path.join(node_line_dir, "query_expansion", "summary.csv")
    assert os.path.exists(summary_path)
    summary_df = load_summary_file(summary_path)
    assert set(summary_df.columns) == {'filename', 'query_expansion_retrieval_f1', 'query_expansion_retrieval_recall',
                                       'module_name', 'module_params', 'execution_time', 'is_best'}
    assert len(summary_df) == 2
    assert summary_df['filename'][0] == "0.parquet"
    assert summary_df['module_name'][0] == "query_decompose"
    assert summary_df['module_params'][0] == {'generator_module_type': 'llama_index_llm', 'llm': "mock", 'batch': 7}
    assert summary_df['execution_time'][0] > 0
    assert summary_df['is_best'][0] == True or summary_df['is_best'][0] == False  # is_best is np.bool_
    # test the best file is saved properly
    best_filename = summary_df[summary_df['is_best']]['filename'].values[0]
    best_path = os.path.join(node_line_dir, "query_expansion", f"best_{best_filename}")
    assert os.path.exists(best_path)
    best_df = pd.read_parquet(best_path)
    assert all([expect_column in best_df.columns for expect_column in expect_columns])


def test_run_query_expansion_node(node_line_dir):
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    previous_result = pd.read_parquet(qa_path)

    modules = [query_decompose, hyde]
    module_params = [{'generator_module_type': 'llama_index_llm', 'llm': "mock", 'batch': 7},
                     {'generator_module_type': 'llama_index_llm', 'llm': "mock"}]
    strategies = {
        'metrics': metrics,
        'speed_threshold': 5,
        'top_k': 4,
        'strategy': 'rank',
        'retrieval_modules': [{'module_type': 'bm25', 'bm25_tokenizer': 'gpt2'}],
    }
    best_result = run_query_expansion_node(modules, module_params, previous_result, node_line_dir, strategies)
    base_query_expansion_test(best_result, node_line_dir)


def test_run_query_expansion_node_default(node_line_dir):
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    previous_result = pd.read_parquet(qa_path)

    modules = [query_decompose, hyde]
    module_params = [{'generator_module_type': 'llama_index_llm', 'llm': "mock", 'batch': 7},
                     {'generator_module_type': 'llama_index_llm', 'llm': "mock"}]
    strategies = {
        'metrics': metrics
    }
    best_result = run_query_expansion_node(modules, module_params, previous_result, node_line_dir, strategies)
    base_query_expansion_test(best_result, node_line_dir)


def test_run_query_expansion_one_module(node_line_dir):
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    previous_result = pd.read_parquet(qa_path)

    modules = [query_decompose]
    module_params = [{'generator_module_type': 'llama_index_llm', 'llm': "mock"}]
    strategies = {
        'metrics': metrics
    }
    best_result = run_query_expansion_node(modules, module_params, previous_result, node_line_dir, strategies)
    assert set(best_result.columns) == {
        'qid', 'query', 'generation_gt', 'retrieval_gt', 'queries'  # automatically skip evaluation
    }
    summary_filepath = os.path.join(node_line_dir, "query_expansion", "summary.csv")
    assert os.path.exists(summary_filepath)
    summary_df = load_summary_file(summary_filepath)
    assert set(summary_df) == {
        'filename', 'module_name', 'module_params', 'execution_time', 'is_best'
    }
    best_filepath = os.path.join(node_line_dir, "query_expansion", f"best_{summary_df['filename'].values[0]}")
    assert os.path.exists(best_filepath)


async def mock_acomplete(self, messages, **kwargs):
    return CompletionResponse(text=messages)


@patch.object(OpenAI, "acomplete", mock_acomplete)
def test_run_query_expansion_no_generator(node_line_dir):
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    previous_result = pd.read_parquet(qa_path)

    modules = [query_decompose]
    module_params = [{}]
    strategies = {
        'metrics': metrics
    }
    best_result = run_query_expansion_node(modules, module_params, previous_result, node_line_dir, strategies)
    assert set(best_result.columns) == {
        'qid', 'query', 'generation_gt', 'retrieval_gt', 'queries'  # automatically skip evaluation
    }
    summary_filepath = os.path.join(node_line_dir, "query_expansion", "summary.csv")
    assert os.path.exists(summary_filepath)
    summary_df = load_summary_file(summary_filepath)
    assert set(summary_df) == {
        'filename', 'module_name', 'module_params', 'execution_time', 'is_best'
    }
    best_filepath = os.path.join(node_line_dir, "query_expansion", f"best_{summary_df['filename'].values[0]}")
    assert os.path.exists(best_filepath)



from typing import List, Callable, Dict

import pandas as pd

from autorag.nodes.queryexpansion.base import query_expansion_node

decompose_prompt = """Decompose a question in self-contained sub-questions. Use \"The question needs no decomposition\" when no decomposition is needed.

    Example 1:

    Question: Is Hamlet more common on IMDB than Comedy of Errors?
    Decompositions: 
    1: How many listings of Hamlet are there on IMDB?
    2: How many listing of Comedy of Errors is there on IMDB?

    Example 2:

    Question: Are birds important to badminton?

    Decompositions:
    The question needs no decomposition

    Example 3:

    Question: Is it legal for a licensed child driving Mercedes-Benz to be employed in US?

    Decompositions:
    1: What is the minimum driving age in the US?
    2: What is the minimum age for someone to be employed in the US?

    Example 4:

    Question: Are all cucumbers the same texture?

    Decompositions:
    The question needs no decomposition

    Example 5:

    Question: Hydrogen's atomic number squared exceeds number of Spice Girls?

    Decompositions:
    1: What is the atomic number of hydrogen?
    2: How many Spice Girls are there?

    Example 6:

    Question: {question}

    Decompositions:
    """


@query_expansion_node
def query_decompose(queries: List[str],
                    generator_func: Callable,
                    generator_params: Dict,
                    prompt: str = decompose_prompt) -> List[List[str]]:
    """
    decompose query to little piece of questions.
    :param queries: List[str], queries to decompose.
    :param generator_func: Callable, generator functions.
    :param generator_params: Dict, generator parameters.
    :param prompt: str, prompt to use for query decomposition.
        default prompt comes from Visconde's StrategyQA few-shot prompt.
    :return: List[List[str]], list of decomposed query. Return input query if query is not decomposable.
    """
    full_prompts = []
    for query in queries:
        if bool(prompt):
            full_prompt = f"prompt: {prompt}\n\n question: {query}"
        else:
            full_prompt = decompose_prompt.format(question=query)
        full_prompts.append(full_prompt)
    input_df = pd.DataFrame({"prompts": full_prompts})
    result_df = generator_func(project_dir=None, previous_result=input_df, **generator_params)
    answers = result_df['generated_texts'].tolist()
    results = list(map(lambda x: get_query_decompose(x[0], x[1]), zip(queries, answers)))
    return results


def get_query_decompose(query: str, answer: str) -> List[str]:
    """
    decompose query to little piece of questions.
    :param query: str, query to decompose.
    :param answer: str, answer from query_decompose function.
    :return: List[str], list of a decomposed query. Return input query if query is not decomposable.
    """
    if answer.lower() == "the question needs no decomposition":
        return [query]
    try:
        lines = [line.strip() for line in answer.splitlines() if line.strip()]
        if lines[0].startswith("Decompositions:"):
            lines.pop(0)
        questions = [line.split(':', 1)[1].strip() for line in lines if ':' in line]
        if not questions:
            return [query]
        return questions
    except:
        return [query]



from typing import List, Tuple, Optional

import numpy as np
import torch.cuda

from autorag import embedding_models
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.nodes.passagefilter.base import passage_filter_node
from autorag.nodes.passagefilter.similarity_threshold_cutoff import embedding_query_content


@passage_filter_node
def similarity_percentile_cutoff(queries: List[str], contents_list: List[List[str]],
                                 scores_list: List[List[float]], ids_list: List[List[str]],
                                 percentile: float, embedding_model: Optional[str] = None,
                                 batch: int = 128,
                                 ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Re-calculate each content's similarity with the query and filter out the contents that are below the content's
    length times percentile. If This is a filter and does not override scores. The output of scores is not coming from
    query-content similarity.
    If the value of content's length times percentile is less than 1, keep the only one highest similarity content.

    :param queries: The list of queries to use for filtering
    :param contents_list: The list of lists of contents to filter
    :param scores_list: The list of lists of scores retrieved
    :param ids_list: The list of lists of ids retrieved
    :param percentile: The percentile to cut off
    :param embedding_model: The embedding model to use for calculating similarity
        Default is OpenAIEmbedding.
    :param batch: The number of queries to be processed in a batch
        Default is 128.
    :return: Tuple of lists containing the filtered contents, ids, and scores
    """
    if embedding_model is None:
        embedding_model = embedding_models['openai']
    else:
        embedding_model = embedding_models[embedding_model]

    query_embeddings, content_embeddings = embedding_query_content(queries, contents_list, embedding_model,
                                                                   batch)

    results = list(map(lambda x: similarity_percentile_cutoff_pure(x[0], x[1], x[2], x[3], x[4], percentile),
                       zip(query_embeddings, content_embeddings, contents_list, ids_list, scores_list)))

    remain_content_list = list(map(lambda x: x[0], results))
    remain_ids_list = list(map(lambda x: x[1], results))
    remain_scores_list = list(map(lambda x: x[2], results))

    del embedding_model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return remain_content_list, remain_ids_list, remain_scores_list


def similarity_percentile_cutoff_pure(query_embedding: str,
                                      content_embeddings: List[List[float]],
                                      content_list: List[str],
                                      ids_list: List[str],
                                      scores_list: List[float],
                                      percentile: float) -> Tuple[List[str], List[str], List[float]]:
    """
    Return tuple of lists containing the filtered contents, ids, and scores

    :param query_embedding: Query embedding
    :param content_embeddings: Each content embedding
    :param content_list: Each content
    :param ids_list: Each id
    :param scores_list: Each score
    :param percentile: The percentile to cut off
    :return: Tuple of lists containing the filtered contents, ids, and scores
    """
    num_top_k = int(len(content_embeddings) * percentile)

    if num_top_k == 0:
        num_top_k = 1

    similarities = np.array(list(map(lambda x: calculate_cosine_similarity(query_embedding, x),
                                     content_embeddings))).tolist()

    content_id_score_similarity = list(zip(ids_list, content_list, scores_list, similarities))

    sorted_content_id_score_similarity = sorted(content_id_score_similarity, key=lambda x: x[3], reverse=True)[
                                         :num_top_k]

    content_result, id_result, score_result, _ = zip(*sorted_content_id_score_similarity)
    return list(content_result), list(id_result), list(score_result)



import logging
from typing import List, Tuple

import pandas as pd
import torch
from tqdm import tqdm
from transformers import T5Tokenizer, T5ForConditionalGeneration

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import select_top_k, sort_by_scores

logger = logging.getLogger("AutoRAG")


@passage_reranker_node
def upr(queries: List[str], contents_list: List[List[str]],
        scores_list: List[List[float]], ids_list: List[List[str]],
        top_k: int, use_bf16: bool = False,
        prefix_prompt: str = "Passage: ",
        suffix_prompt: str = "Please write a question based on this passage.") \
        -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using UPR.
    UPR is a reranker based on UPR (https://github.com/DevSinghSachan/unsupervised-passage-reranking).
    The language model will make a question based on the passage and rerank the passages by the likelihood of the question.
    The default model is t5-large.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param use_bf16: Whether to use bfloat16 for the model. Default is False.
    :param prefix_prompt: The prefix prompt for the language model that generates question for reranking.
        Default is "Passage: ".
        The prefix prompt serves as the initial context or instruction for the language model.
        It sets the stage for what is expected in the output
    :param suffix_prompt: The suffix prompt for the language model that generates question for reranking.
        Default is "Please write a question based on this passage.".
        The suffix prompt provides a cue or a closing instruction to the language model,
            signaling how to conclude the generated text or what format to follow at the end.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    tqdm.pandas()
    df = pd.DataFrame({
        'query': queries,
        'contents': contents_list,
        'ids': ids_list,
    })

    scorer = UPRScorer(suffix_prompt=suffix_prompt, prefix_prompt=prefix_prompt, use_bf16=use_bf16)

    df['scores'] = df.progress_apply(lambda row: scorer.compute(query=row['query'], contents=row['contents']),
                                     axis=1)
    del scorer
    df[['contents', 'ids', 'scores']] = df.apply(lambda x: sort_by_scores(x, reverse=False), axis=1,
                                                 result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)
    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


class UPRScorer:
    def __init__(self, suffix_prompt: str, prefix_prompt: str,
                 use_bf16: bool = False):
        model_name = "t5-large"
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.tokenizer = T5Tokenizer.from_pretrained(model_name)
        self.model = T5ForConditionalGeneration.from_pretrained(model_name,
                                                                torch_dtype=torch.bfloat16
                                                                if use_bf16 else torch.float32).to(self.device)
        self.suffix_prompt = suffix_prompt
        self.prefix_prompt = prefix_prompt

    def compute(self, query: str, contents: List[str]) -> List[float]:
        query_token = self.tokenizer(query, max_length=128, truncation=True, return_tensors='pt')
        prompts = list(map(lambda content: f"{self.prefix_prompt} {content} {self.suffix_prompt}", contents))
        prompt_token_outputs = self.tokenizer(prompts, padding='longest',
                                              max_length=512,
                                              pad_to_multiple_of=8,
                                              truncation=True,
                                              return_tensors='pt')

        query_input_ids = torch.repeat_interleave(query_token['input_ids'], len(contents),
                                                  dim=0).to(self.device)

        with torch.no_grad():
            logits = self.model(input_ids=prompt_token_outputs['input_ids'].to(self.device),
                                attention_mask=prompt_token_outputs['attention_mask'].to(self.device),
                                labels=query_input_ids).logits
        log_softmax = torch.nn.functional.log_softmax(logits, dim=-1)
        nll = -log_softmax.gather(2, query_input_ids.unsqueeze(2)).squeeze(2)
        avg_nll = torch.sum(nll, dim=1)
        return avg_nll.tolist()

    def __del__(self):
        del self.model
        del self.tokenizer
        if torch.cuda.is_available():
            torch.cuda.empty_cache()



from autorag.nodes.passageaugmenter import pass_passage_augmenter

from tests.autorag.nodes.passageaugmenter.test_base_passage_augmenter import project_dir, previous_result


def test_pass_passage_augmenter():
    result_df = pass_passage_augmenter(project_dir=project_dir, previous_result=previous_result, top_k=2)
    assert all([column_name in result_df.columns for column_name in
                ["retrieved_contents", "retrieved_ids", "retrieve_scores"]])
    assert result_df["retrieved_contents"].tolist() == previous_result["retrieved_contents"].tolist()
    assert result_df["retrieved_ids"].tolist() == previous_result["retrieved_ids"].tolist()
    assert result_df["retrieve_scores"].tolist() == previous_result["retrieve_scores"].tolist()



from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagefilter.base import passage_filter_node
from autorag.utils.util import sort_by_scores, select_top_k


@passage_filter_node
def percentile_cutoff(queries: List[str], contents_list: List[List[str]],
                      scores_list: List[List[float]], ids_list: List[List[str]],
                      percentile: float, reverse: bool = False,
                      ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Filter out the contents that are below the content's length times percentile.
    If This is a filter and does not override scores.
    If the value of content's length times percentile is less than 1, keep the only one highest similarity content.

    :param queries: The list of queries to use for filtering
    :param contents_list: The list of lists of contents to filter
    :param scores_list: The list of lists of scores retrieved
    :param ids_list: The list of lists of ids retrieved
    :param percentile: The percentile to cut off
    :param reverse: If True, the lower the score, the better
        Default is False.
    :return: Tuple of lists containing the filtered contents, ids, and scores
    """
    num_top_k = max(1, int(len(scores_list[0]) * percentile))

    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': scores_list,
    })

    reverse = not reverse
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand', reverse=reverse)
    results = select_top_k(df, ['contents', 'ids', 'scores'], num_top_k)

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()



import os
import pathlib

from datetime import datetime

import pandas as pd

from autorag.nodes.retrieval.base import evenly_distribute_passages

queries = [
    ["What is Visconde structure?", "What are Visconde structure?"],
    ["What is the structure of StrategyQA dataset in this paper?"],
    ["What's your favorite source of RAG framework?",
     "What is your source of RAG framework?",
     "Is RAG framework have source?"],
]

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
project_dir = os.path.join(root_dir, "resources", "sample_project")
qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
corpus_data = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
previous_result = qa_data.sample(5)

doc_id = ["doc1", "doc2", "doc3", "doc4", "doc5"]
contents = ["This is a test document 1.", "This is a test document 2.", "This is a test document 3.",
            "This is a test document 4.", "This is a test document 5."]
metadata = [{'datetime': datetime.now()} for _ in range(5)]
corpus_df = pd.DataFrame({"doc_id": doc_id, "contents": contents, "metadata": metadata})


def base_retrieval_test(id_result, score_result, top_k):
    assert len(id_result) == len(score_result) == 3
    for id_list, score_list in zip(id_result, score_result):
        assert isinstance(id_list, list)
        assert isinstance(score_list, list)
        assert len(id_list) == len(score_list) == top_k
        for _id, score in zip(id_list, score_list):
            assert isinstance(_id, str)
            assert isinstance(score, float)
        for i in range(1, len(score_list)):
            assert score_list[i - 1] >= score_list[i]


def base_retrieval_node_test(result_df):
    contents = result_df["retrieved_contents"].tolist()
    ids = result_df["retrieved_ids"].tolist()
    scores = result_df["retrieve_scores"].tolist()
    assert len(contents) == len(ids) == len(scores) == 5
    assert len(contents[0]) == len(ids[0]) == len(scores[0]) == 4
    # id is matching with corpus.parquet
    for content_list, id_list, score_list in zip(contents, ids, scores):
        for i, (content, _id, score) in enumerate(zip(content_list, id_list, score_list)):
            assert isinstance(content, str)
            assert isinstance(_id, str)
            assert isinstance(score, float)
            assert _id in corpus_data["doc_id"].tolist()
            assert content == corpus_data[corpus_data["doc_id"] == _id]["contents"].values[0]
            if i >= 1:
                assert score_list[i - 1] >= score_list[i]


def test_evenly_distribute_passages():
    ids = [[f'test-{i}-{j}' for i in range(10)] for j in range(3)]
    scores = [[i for i in range(10)] for _ in range(3)]
    top_k = 10
    new_ids, new_scores = evenly_distribute_passages(ids, scores, top_k)
    assert len(new_ids) == top_k
    assert len(new_scores) == top_k
    assert new_scores == [0, 1, 2, 3, 0, 1, 2, 0, 1, 2]



import os
import pathlib
from typing import Callable, List, Dict, Union

import pandas as pd

from autorag.evaluation import evaluate_generation
from autorag.evaluation.util import cast_metrics
from autorag.strategy import measure_speed, filter_by_threshold, select_best


def run_generator_node(modules: List[Callable],
                       module_params: List[Dict],
                       previous_result: pd.DataFrame,
                       node_line_dir: str,
                       strategies: Dict,
                       ) -> pd.DataFrame:
    """
    Run evaluation and select the best module among generator node results.
    And save the results and summary to generator node directory.

    :param modules: Generator modules to run.
    :param module_params: Generator module parameters.
        Including node parameters, which is used for every module in this node.
    :param previous_result: Previous result dataframe.
        Could be prompt maker node's result.
    :param node_line_dir: This node line's directory.
    :param strategies: Strategies for generator node.
    :return: The best result dataframe.
        It contains previous result columns and generator node's result columns.
    """
    if not os.path.exists(node_line_dir):
        os.makedirs(node_line_dir)
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    node_dir = os.path.join(node_line_dir, "generator")  # node name
    if not os.path.exists(node_dir):
        os.makedirs(node_dir)
    qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"), engine='pyarrow')
    if 'generation_gt' not in qa_data.columns:
        raise ValueError("You must have 'generation_gt' column in qa.parquet.")
    generation_gt = list(map(lambda x: x.tolist(), qa_data['generation_gt'].tolist()))

    results, execution_times = zip(*map(lambda x: measure_speed(
        x[0], project_dir=project_dir, previous_result=previous_result, **x[1]),
                                        zip(modules, module_params)))
    average_times = list(map(lambda x: x / len(results[0]), execution_times))

    # get average token usage
    token_usages = list(map(lambda x: x['generated_tokens'].apply(len).mean(), results))

    metric_names, metric_params = cast_metrics(strategies.get('metrics'))
    if metric_names is None or len(metric_names) <= 0:
        raise ValueError("You must at least one metrics for generator evaluation.")
    results = list(
        map(lambda result: evaluate_generator_node(result, generation_gt, strategies.get('metrics')), results))

    # save results to folder
    filepaths = list(map(lambda x: os.path.join(node_dir, f'{x}.parquet'), range(len(modules))))
    list(map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths)))  # execute save to parquet
    filenames = list(map(lambda x: os.path.basename(x), filepaths))

    summary_df = pd.DataFrame({
        'filename': filenames,
        'module_name': list(map(lambda module: module.__name__, modules)),
        'module_params': module_params,
        'execution_time': average_times,
        'average_output_token': token_usages,
        **{metric: list(map(lambda x: x[metric].mean(), results)) for metric in metric_names}
    })

    # filter by strategies
    if strategies.get('speed_threshold') is not None:
        results, filenames = filter_by_threshold(results, average_times, strategies['speed_threshold'], filenames)
    if strategies.get('token_threshold') is not None:
        results, filenames = filter_by_threshold(results, token_usages, strategies['token_threshold'], filenames)
    selected_result, selected_filename = select_best(results, metric_names, filenames,
                                                     strategies.get('strategy', 'mean'))
    best_result = pd.concat([previous_result, selected_result], axis=1)

    # add 'is_best' column at summary file
    summary_df['is_best'] = summary_df['filename'] == selected_filename

    # save files
    summary_df.to_csv(os.path.join(node_dir, "summary.csv"), index=False)
    best_result.to_parquet(os.path.join(node_dir, f"best_{os.path.splitext(selected_filename)[0]}.parquet"),
                           index=False)
    return best_result


def evaluate_generator_node(result_df: pd.DataFrame, generation_gt, metrics: Union[List[str], List[Dict]]):
    @evaluate_generation(generation_gt=generation_gt, metrics=metrics)
    def evaluate_generation_module(df: pd.DataFrame):
        return df['generated_texts'].tolist(), df['generated_tokens'].tolist(), df['generated_log_probs'].tolist()

    return evaluate_generation_module(result_df)



import pytest

from autorag.evaluation.metric.retrieval_contents import single_token_f1, retrieval_token_f1, retrieval_token_precision, \
    retrieval_token_recall

gt = [
    ['Enough for drinking water', 'Just looking for a water bottle'],
    ['Do you want to buy some?'],
    [''],
    []
]
pred = [
    ['Enough for mixing water', 'I want to do a nothing', 'Just looking is a very healthy'],
    ['Do you want to buy some?', 'I want to buy some', 'I want to buy some water'],
    ['Who is son? He is great player in the world'],
    ['i love havertz', 'i love kai havertz']
]


def test_single_token_f1():
    precision, recall, f1 = single_token_f1(gt[0][0], pred[0][0])
    assert precision == 0.75
    assert recall == 0.75
    assert f1 == 0.75

    precision, recall, f1 = single_token_f1(gt[0][1], pred[0][2])
    assert precision == 0.4
    assert recall == 0.4
    assert f1 == pytest.approx(0.4)


def test_retrieval_token_f1():
    f1 = retrieval_token_f1.__wrapped__(gt[0], pred[0])
    assert f1 == pytest.approx(0.38333, rel=0.001)

    f1 = retrieval_token_f1.__wrapped__(gt[1], pred[1])
    assert f1 == pytest.approx(0.797979, rel=0.001)

    result_f1 = retrieval_token_f1(gt_contents=gt, pred_contents=pred)
    assert result_f1 == pytest.approx([0.38333, 0.797979, None, None], rel=0.001)


def test_retrieval_token_precision():
    result_precision = retrieval_token_precision(gt_contents=gt, pred_contents=pred)
    assert result_precision == pytest.approx([0.383333, 0.8222222, None, None], rel=0.001)


def test_retrieval_token_recall():
    result_recall = retrieval_token_recall(gt_contents=gt, pred_contents=pred)
    assert result_recall == pytest.approx([0.383333, 0.777777, None, None], rel=0.001)



import os.path
import tempfile

import pytest

from autorag.nodes.promptmaker import window_replacement
from tests.autorag.nodes.promptmaker.test_prompt_maker_base import (prompt, queries, retrieved_contents,
                                                                    retrieved_metadata, previous_result, corpus_df)


@pytest.fixture
def pseudo_project_dir():
    with tempfile.TemporaryDirectory() as project_dir:
        data_dir = os.path.join(project_dir, 'data')
        os.makedirs(data_dir)
        corpus_df.to_parquet(os.path.join(data_dir, 'corpus.parquet'))
        yield project_dir


def test_window_replacement():
    window_replacement_original = window_replacement.__wrapped__
    result_prompts = window_replacement_original(prompt, queries, retrieved_contents, retrieved_metadata)
    assert len(result_prompts) == 2
    assert isinstance(result_prompts, list)
    assert result_prompts[
               0] == "Answer this question: What is the capital of Japan? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"
    assert result_prompts[
               1] == "Answer this question: What is the capital of China? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"


def test_window_replacement_node(pseudo_project_dir):
    result = window_replacement(project_dir=pseudo_project_dir,
                                previous_result=previous_result,
                                prompt=prompt)
    assert len(result) == 2
    assert result.columns == ["prompts"]
    assert result['prompts'][
               0] == "Answer this question: What is the capital of Japan? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"
    assert result['prompts'][
               1] == "Answer this question: What is the capital of China? \n\n havertz arsenal doosan minji naeun gaeun lets go\n\nhavertz arsenal doosan minji naeun gaeun lets go"



import os
import pathlib
import click

from datasets import load_dataset


@click.command()
@click.option('--save_path', type=str, default=pathlib.PurePath(__file__).parent, help='Path to save sample eli5 dataset.')
def load_eli5_dataset(save_path):
    # set file path
    file_path = "MarkrAI/eli5_sample_autorag"

    # load dataset
    corpus_dataset = load_dataset(file_path, "corpus")['train'].to_pandas()
    qa_train_dataset = load_dataset(file_path, "qa")['train'].to_pandas()
    qa_test_dataset = load_dataset(file_path, "qa")['test'].to_pandas()

    # save data
    if os.path.exists(os.path.join(save_path, "corpus.parquet")) is True:
        raise ValueError("corpus.parquet already exists")
    if os.path.exists(os.path.join(save_path, "qa.parquet")) is True:
        raise ValueError("qa.parquet already exists")
    corpus_dataset.to_parquet(os.path.join(save_path, "corpus.parquet"))
    qa_train_dataset.to_parquet(os.path.join(save_path, "qa_train.parquet"))
    qa_test_dataset.to_parquet(os.path.join(save_path, "qa_test.parquet"))


if __name__ == '__main__':
    load_eli5_dataset()



import numpy as np
import pytest

from autorag.evaluation.metric import (retrieval_f1, retrieval_precision, retrieval_recall, retrieval_ndcg,
                                       retrieval_mrr,
                                       retrieval_map)

retrieval_gt = [
    [['test-1', 'test-2'], ['test-3']],
    [['test-4', 'test-5'], ['test-6', 'test-7'], ['test-8']],
    [['test-9', 'test-10']],
    [['test-11'], ['test-12'], ['test-13']],
    [['test-14']],
    [[]],
    [['']],
    [['test-15']]
]

pred = [
    ['test-1', 'pred-1', 'test-2', 'pred-3'],  # recall: 0.5, precision: 0.5, f1: 0.5
    ['test-6', 'pred-5', 'pred-6', 'pred-7'],  # recall: 1/3, precision: 0.25, f1: 2/7
    ['test-9', 'pred-0', 'pred-8', 'pred-9'],  # recall: 1.0, precision: 0.25, f1: 2/5
    ['test-13', 'test-12', 'pred-10', 'pred-11'],  # recall: 2/3, precision: 0.5, f1: 4/7
    ['test-14', 'pred-12'],  # recall: 1.0, precision: 0.5, f1: 2/3
    ['pred-13'],  # retrieval_gt is empty so not counted
    ['pred-14'],  # retrieval_gt is empty so not counted
    ['pred-15', 'pred-16', 'test-15']  # recall:1, precision: 1/3, f1: 0.5
]


def test_retrieval_f1():
    solution = [0.5, 2 / 7, 2 / 5, 4 / 7, 2 / 3, None, None, 0.5]
    result = retrieval_f1(retrieval_gt=retrieval_gt, pred_ids=pred)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)


def test_numpy_retrieval_metric():
    retrieval_gt_np = [[np.array(['test-1', 'test-4'])], np.array([['test-2']])]
    pred_np = np.array([['test-2', 'test-3', 'test-1'], ['test-5', 'test-6', 'test-8']])
    solution = [1.0, 0.0]
    result = retrieval_recall(retrieval_gt=retrieval_gt_np, pred_ids=pred_np)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_recall():
    solution = [0.5, 1 / 3, 1, 2 / 3, 1, None, None, 1]
    result = retrieval_recall(retrieval_gt=retrieval_gt, pred_ids=pred)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_precision():
    solution = [0.5, 0.25, 0.25, 0.5, 0.5, None, None, 1 / 3]
    result = retrieval_precision(retrieval_gt=retrieval_gt, pred_ids=pred)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_ndcg():
    solution = [0.7039180890341347, 0.3903800499921017, 0.6131471927654584, 0.7653606369886217, 1, None, None, 0.5]
    result = retrieval_ndcg(retrieval_gt=retrieval_gt, pred_ids=pred)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_mrr():
    solution = [1 / 2, 1 / 3, 1, 1 / 2, 1, None, None, 1 / 3]
    result = retrieval_mrr(retrieval_gt=retrieval_gt, pred_ids=pred)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)


def test_retrieval_map():
    solution = [5 / 12, 1 / 3, 1, 1 / 2, 1, None, None, 1 / 3]
    result = retrieval_map(retrieval_gt=retrieval_gt, pred_ids=pred)
    for gt, res in zip(solution, result):
        assert gt == pytest.approx(res, rel=1e-4)



import logging
from typing import List, Dict

from autorag.nodes.promptmaker.base import prompt_maker_node

logger = logging.getLogger("AutoRAG")


@prompt_maker_node
def window_replacement(prompt: str, queries: List[str],
                       retrieved_contents: List[List[str]],
                       retrieved_metadata: List[List[Dict]]) -> List[str]:
    """
    Replace retrieved_contents with window to create a Prompt
    (only available for corpus chunked with Sentence window method)
    You must type a prompt or prompt list at config yaml file like this:

    .. Code:: yaml
    nodes:
    - node_type: prompt_maker
      modules:
      - module_type: window_replacement
        prompt: [Answer this question: {query} \n\n {retrieved_contents},
        Read the passages carefully and answer this question: {query} \n\n Passages: {retrieved_contents}]

    :param prompt: A prompt string.
    :param queries: List of query strings.
    :param retrieved_contents: List of retrieved contents.
    :param retrieved_metadata: List of retrieved metadata.
    :return: Prompts that made by window_replacement.
    """

    def window_replacement_row(_prompt: str, _query: str, _retrieved_contents, _retrieved_metadata: List[Dict]) -> str:
        window_list = []
        for content, metadata in zip(_retrieved_contents, _retrieved_metadata):
            if 'window' in metadata:
                window_list.append(metadata['window'])
            else:
                window_list.append(content)
                logger.info("Only available for corpus chunked with Sentence window method."
                            "window_replacement will not proceed.")
        contents_str = "\n\n".join(window_list)
        return _prompt.format(query=_query, retrieved_contents=contents_str)

    return list(map(lambda x: window_replacement_row(prompt, x[0], x[1], x[2]),
                    zip(queries, retrieved_contents, retrieved_metadata)))



import asyncio
import itertools
import os
import pathlib
import tempfile
from datetime import datetime, date

import numpy as np
import pandas as pd
import pytest
import tiktoken
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.llms import CompletionResponse
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.utils import fetch_contents
from autorag.utils.util import load_summary_file, result_to_dataframe, \
    make_combinations, explode, replace_value_in_dict, normalize_string, convert_string_to_tuple_in_dict, process_batch, \
    convert_env_in_dict, openai_truncate_by_token, convert_datetime_string, split_dataframe, find_trial_dir, \
    find_node_summary_files, normalize_unicode, dict_to_markdown, dict_to_markdown_table, convert_inputs_to_list, \
    to_list
from tests.mock import MockLLM

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent

summary_df = pd.DataFrame({
    'best_module_name': ['bm25', 'upr', 'gpt-4'],
    'best_module_params': [{'top_k': 50}, {'model': 'llama-2', 'havertz': 'chelsea'}, {'top_p': 0.9}],
})


@pytest.fixture
def module_name():
    return "test_module"


@pytest.fixture
def module_params():
    return {
        "param1": "value1",
        "param2": "value2",
        "param3": "value3",
    }


@pytest.fixture
def summary_path():
    with tempfile.TemporaryDirectory() as tmp_dir:
        summary_path = os.path.join(tmp_dir, "summary.csv")
        summary_df.to_csv(summary_path, index=False)
        yield summary_path


def test_fetch_contents():
    corpus_data_path = os.path.join(root_dir, "resources", "corpus_data_sample.parquet")
    corpus_data = pd.read_parquet(corpus_data_path)
    search_rows = corpus_data.sample(n=10)
    find_contents = fetch_contents(corpus_data, list(map(lambda x: [x], search_rows['doc_id'].tolist())))
    assert len(find_contents) == len(search_rows)
    assert list(itertools.chain.from_iterable(find_contents)) == search_rows['contents'].tolist()

    corpus_data = pd.DataFrame({
        'doc_id': ['doc1', 'doc2', 'doc3'],
        'contents': ['apple', 'banana', 'cherry'],
        'metadata': [{'last_modified_datetime': datetime(2022, 1, 1, 0, 0, 0)},
                     {'last_modified_datetime': datetime(2022, 1, 2, 0, 0, 0)},
                     {'last_modified_datetime': datetime(2022, 1, 3, 0, 0, 0)}]
    })
    find_contents = fetch_contents(corpus_data, [['doc3', 'doc1'], ['doc2']])
    assert find_contents[0] == ['cherry', 'apple']
    assert find_contents[1] == ['banana']

    find_metadatas = fetch_contents(corpus_data, [['doc3', 'doc1'], ['doc2']], 'metadata')
    assert find_metadatas[0] == [{'last_modified_datetime': datetime(2022, 1, 3, 0, 0, 0)},
                                 {'last_modified_datetime': datetime(2022, 1, 1, 0, 0, 0)}]
    assert find_metadatas[1] == [{'last_modified_datetime': datetime(2022, 1, 2, 0, 0, 0)}]

    find_empty = fetch_contents(corpus_data, [[], ['doc2']])
    assert find_empty[0] == [None]
    assert find_empty[1] == ['banana']

    find_blank = fetch_contents(corpus_data, [[''], ['doc2']])
    assert find_blank[0] == [None]
    assert find_blank[1] == ['banana']


def test_load_summary_file(summary_path):
    with pytest.raises(ValueError):
        load_summary_file(summary_path)
    df = load_summary_file(summary_path, ['best_module_params'])
    assert df.equals(summary_df)


def test_load_summary_file_recency_filter():
    df = pd.DataFrame({
        'module_name': ['havertz', 'recency_filter'],
        'module_params': [{'jazz': 'eastsidegunn'},
                          {'threshold': datetime(2022, 1, 3, 0, 1, 3)}],
    })
    with tempfile.NamedTemporaryFile(suffix='.csv') as csv_file:
        df.to_csv(csv_file.name, index=False)
        load_df = load_summary_file(csv_file.name)
        assert load_df.equals(df)


def test_convert_datetime_string():
    datetime_dict = {'threshold': datetime(2022, 1, 3, 0, 0, 3)}
    date_dict = {'threshold': date(2001, 7, 11)}
    result1 = convert_datetime_string(str(datetime_dict))
    result2 = convert_datetime_string(str(date_dict))

    assert result1 == datetime_dict['threshold']
    assert result2 == date_dict['threshold']


def test_result_to_dataframe():
    @result_to_dataframe(['col_1', 'col_2'])
    def func1():
        return [1, 2], [3, 4]

    result1 = func1()
    assert isinstance(result1, pd.DataFrame)
    assert result1.columns.tolist() == ['col_1', 'col_2']
    assert result1['col_1'].tolist() == [1, 2]
    assert result1['col_2'].tolist() == [3, 4]

    @result_to_dataframe(['col_1'])
    def func2():
        return [1, 2, 3]

    result2 = func2()
    assert isinstance(result2, pd.DataFrame)
    assert result2.columns.tolist() == ['col_1']
    assert result2['col_1'].tolist() == [1, 2, 3]


def test_make_combinations():
    target_dict = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': ['value4', 'value5']}
    solution = [
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': 'value4'},
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': 'value5'},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': 'value4'},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': 'value5'}
    ]
    combinations = make_combinations(target_dict)
    assert len(combinations) == len(solution)
    assert all([combination in solution for combination in combinations])

    elem1 = {
        'key5': 'value5',
        'key6': ['value6', 'value7']
    }
    elem2 = {'key7': 'value8'}
    value_of_key_4 = [elem1, elem2]
    target_dict = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': value_of_key_4}
    combinations = make_combinations(target_dict)
    solution = [
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': elem1},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': elem1},
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': elem2},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': elem2},
    ]
    assert len(combinations) == len(solution)
    assert all([combination in solution for combination in combinations])

    target_dict = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': ('value4', 'value5')}
    solution = [
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': ('value4', 'value5')},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': ('value4', 'value5')},
    ]
    combinations = make_combinations(target_dict)
    assert len(combinations) == len(solution)
    assert all([combination in solution for combination in combinations])


def test_explode():
    index_values = ['a', 'b', 'c']
    explode_values = [
        ['apple', 'banana', 'cherry'],
        ['april', 'may'],
        ['alpha'],
    ]
    result_index, result_values = explode(index_values, explode_values)
    assert result_index == ['a', 'a', 'a', 'b', 'b', 'c']
    assert result_values == ['apple', 'banana', 'cherry', 'april', 'may', 'alpha']


def test_replace_value_in_dict():
    target_dict = {
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
    }
    result_dict = replace_value_in_dict(target_dict, 'key1', 'value4')
    assert result_dict == {
        'key1': 'value4',
        'key2': 'value2',
        'key3': 'value3',
    }
    result_dict = replace_value_in_dict(target_dict, 'key4', 'value4')
    assert result_dict == target_dict


def test_normalize_string():
    text = "This IS a TEST Text."
    expected = "this is test text"
    assert normalize_string(text) == expected

    text = "Hello, world! This is a test."
    expected = "hello world this is test"
    assert normalize_string(text) == expected

    text = "The quick brown fox jumps over the lazy dog."
    expected = "quick brown fox jumps over lazy dog"
    assert normalize_string(text) == expected

    text = "This    is      a test    text."
    expected = "this is test text"
    assert normalize_string(text) == expected

    text = "The, QUICK Brown-Fox; jumps over... the LAZY dog!"
    expected = "quick brownfox jumps over lazy dog"
    assert normalize_string(text) == expected


def test_convert_string_to_tuple_in_dict():
    # Example usage
    data = {
        'key1': '(1, \'two\', 3)',
        'key2': ['(4, 5, \'six\')', {'nested_key': '(7, 8, \'nine\')'},
                 {'key4': 'value2'}],
        'key3': {'nested_key2': '(10, \'eleven\', 12)',
                 'nested_key3': 'value1',
                 'nested_key4': {'nested_key5': '(\'thirteen\', 14, 15)'}},
    }
    result = convert_string_to_tuple_in_dict(data)
    assert result == {
        'key1': (1, 'two', 3),
        'key2': [
            (4, 5, 'six'),
            {
                'nested_key': (7, 8, 'nine')
            },
            {
                'key4': 'value2'
            }
        ],
        'key3': {
            'nested_key2': (10, 'eleven', 12),
            'nested_key3': 'value1',
            'nested_key4': {
                'nested_key5': ('thirteen', 14, 15)
            }
        }
    }


def test_convert_env_in_dict():
    os.environ['ENV_VAR1'] = 'value1'
    os.environ['ENV_VAR2'] = 'value2'
    os.environ['ENV_VAR3'] = 'value3'
    data = {
        'key1': 'value1',
        'key2': [
            'value1',
            '${ENV_VAR1}',
        ],
        'key3': '${ENV_VAR2}',
        'key4': {
            'key5': 'value1',
            'key6': '${ENV_VAR3}',
            'key7': [
                'value1',
                '${ENV_VAR4}',
            ]
        },
        'prompt': 'This is a prompt with ${ENV_VAR1} and ${ENV_VAR2}.'
    }
    result = convert_env_in_dict(data)
    assert result == {
        'key1': 'value1',
        'key2': [
            'value1',
            'value1',
        ],
        'key3': 'value2',
        'key4': {
            'key5': 'value1',
            'key6': 'value3',
            'key7': [
                'value1',
                '',
            ]
        },
        'prompt': 'This is a prompt with value1 and value2.'
    }


def test_process_batch():
    prompts = [str(i) for i in range(1000)]
    results = [CompletionResponse(text=prompt) for prompt in prompts]
    mock_llm = MockLLM()

    tasks = [mock_llm.acomplete(prompt) for prompt in prompts]
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(process_batch(tasks, batch_size=64))

    assert result == results


def test_openai_truncate_by_token():
    base_text = "This is a test text."
    t1 = base_text * 5
    t2 = base_text * 2000
    t3 = base_text * 20

    truncated = openai_truncate_by_token([t1, t2, t3], 8192,
                                         'text-embedding-ada-002')
    assert len(truncated) == 3
    assert truncated[0] == base_text * 5
    assert len(truncated[1]) < len(t2)
    assert len(tiktoken.encoding_for_model('text-embedding-ada-002').encode(truncated[1])) == 8192
    assert len(truncated[2]) == len(t3)


def test_split_dataframe():
    df = pd.DataFrame({'a': list(range(10)), 'b': list(range(10, 20))})

    df_list_1 = split_dataframe(df, chunk_size=5)
    assert len(df_list_1) == 2
    assert len(df_list_1[0]) == 5
    assert pd.DataFrame({'a': list(range(5)), 'b': list(range(10, 15))}).equals(df_list_1[0])

    df_list_2 = split_dataframe(df, chunk_size=3)
    assert len(df_list_2) == 4
    assert len(df_list_2[0]) == 3
    assert len(df_list_2[-1]) == 1
    assert pd.DataFrame({'a': list(range(3)), 'b': list(range(10, 13))}).equals(df_list_2[0])


def test_find_trial_dir():
    project_dir = os.path.join(root_dir, "resources", "result_project")
    trial_dirs = find_trial_dir(project_dir)

    assert len(trial_dirs) == 4
    assert all(isinstance(int(os.path.basename(path)), int) for path in trial_dirs)


def test_find_node_summary_files():
    trial_dir = os.path.join(root_dir, "resources", "result_project", "2")
    node_summary_paths = find_node_summary_files(trial_dir)

    assert len(node_summary_paths) == 4
    assert all(os.path.basename(path) == 'summary.csv' for path in node_summary_paths)


def test_normalize_unicode():
    str1 = "전국보행자전용도로표준데이터"
    str2 = "전국보행자전용도로표준데이터"
    assert len(str1) == 14
    assert len(str2) == 34
    assert str1 != str2

    new_str1 = normalize_unicode(str1)
    new_str2 = normalize_unicode(str2)

    assert len(new_str1) == 14
    assert len(new_str2) == 14
    assert new_str1 == new_str2


def test_dict_to_markdown():
    data = {
        "Title": "Sample Document",
        "Author": "John Doe",
        "Content": {
            "Introduction": "This is the introduction.",
            "Body": [
                "First point",
                "Second point",
                {"Subsection": "Details about the second point"}
            ],
            "Conclusion": "This is the conclusion."
        }
    }
    markdown_text = dict_to_markdown(data)
    result_text = f"""# Title
Sample Document
# Author
John Doe
# Content
## Introduction
This is the introduction.
## Body
- First point
- Second point
### Subsection
Details about the second point
## Conclusion
This is the conclusion.
"""
    assert result_text == markdown_text


def test_dict_to_markdown_table():
    data = {"key1": "value1", "key2": "value2"}
    result = dict_to_markdown_table(data, 'havertz', 'william')
    result_text = """| havertz | william |
| :---: | :-----: |
| key1 | value1 |
| key2 | value2 |
"""
    assert result == result_text


@convert_inputs_to_list
def convert_inputs_to_list_function(int_type, str_type, iterable_type, iterable_type2):
    assert isinstance(int_type, int)
    assert isinstance(str_type, str)
    assert isinstance(iterable_type, list)
    assert isinstance(iterable_type2, list)


def test_convert_inputs_to_list():
    convert_inputs_to_list_function(1, 'jax', (2, 3), (5, 6, [4, 66]))
    convert_inputs_to_list_function(1, 'jax', np.array([3, 4]), [pd.Series([12, 13]), 14])
    convert_inputs_to_list_function(4, 'jax', pd.Series([7, 8, 9]), np.array([[3, 4], [4, 5]]))


def test_to_list():
    embedding_model = OpenAIEmbedding()
    new_model = to_list(embedding_model)
    assert isinstance(new_model, BaseEmbedding)



import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.evaluation.metric import bleu, meteor, rouge, sem_score, g_eval, bert_score
from tests.delete_tests import is_github_action

generation_gts = [
    ['The dog had bit the man.', 'The man had bitten the dog.'],
    ['I want to be a artist, but I end up to be a programmer.'],
    ['To be a artist these days, you can overcome by AI.',
     'To be a programmer these days, you can overcome by AI.',
     'To be a lawyer these days, you can overcome by AI.'],
]
generations = [
    'The dog bit the man.',
    'It really like to be a programmer, but I think artist is my passion.',
    'To be a artist these days, you can overcome by AI.',
]

ko_generation_gts = [
    ['개가 남자를 물었다.', '남자가 개를 물었다.'],
    ['나는 예술가가 되고 싶었지만, 결국 프로그래머가 되었다.'],
    ['요즘 예술가가 되려면, AI를 이겨야 한다.',
     '요즘 프로그래머가 되려면, AI를 이겨야 한다.',
     '요즘 변호사가 되려면, AI를 이겨야 한다.'],
]

ko_generations = [
    '개가 남자를 물었다.',
    '나는 정말이지 예술가가 되고 싶었지만, 결국 프로그래머가 되었다.',
    '요즘 세상에서는 예술가가 되려면, AI를 이겨야 한다.',
]


def base_test_generation_metrics(func, solution, **kwargs):
    scores = func(generation_gt=generation_gts, generations=generations, **kwargs)
    assert len(scores) == len(generation_gts)
    assert all(isinstance(score, float) for score in scores)
    assert all(list(map(lambda x: x[0] == pytest.approx(x[1], 0.001),
                        zip(scores, solution))))


def ko_base_test_generation_metrics(func, solution, **kwargs):
    scores = func(generation_gt=ko_generation_gts, generations=ko_generations, **kwargs)
    assert len(scores) == len(ko_generation_gts)
    assert all(isinstance(score, float) for score in scores)
    assert all(list(map(lambda x: x[0] == pytest.approx(x[1], 0.001),
                        zip(scores, solution))))

def test_bleu():
    base_test_generation_metrics(bleu, [51.1507, 23.5783, 100.0], lowercase=True)


def test_meteor():
    base_test_generation_metrics(meteor, [0.454033, 0.2985435, 0.64077828], alpha=0.85, beta=0.2, gamma=0.6)


def test_rouge():
    base_test_generation_metrics(rouge, [0.909, 0.35714, 1.0])


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions. It use local model")
def test_sem_score():
    base_test_generation_metrics(sem_score, [0.9005998, 0.7952, 1.0])


def test_sem_score_other_model():
    base_test_generation_metrics(sem_score, [0.9888, 0.9394, 1.0], embedding_model=OpenAIEmbedding())


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_g_eval_fluency():
    base_test_generation_metrics(g_eval, [3.0, 2.0, 2.0], metrics=['fluency'], model='gpt-3.5-turbo')


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_g_eval_full():
    base_test_generation_metrics(g_eval, [4.25, 2.5, 4.25], model='gpt-3.5-turbo')


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_bert_score_en():
    base_test_generation_metrics(bert_score, [0.981902, 0.93164, 1.0], n_threads=8)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_bert_score_ko():
    ko_base_test_generation_metrics(bert_score, [1.0, 0.965312, 0.96309], lang='ko')



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import copy

import torch
from torch import nn
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from transformers.modeling_outputs import SequenceClassifierOutput
from transformers.models.t5.modeling_t5 import T5Config, T5PreTrainedModel, T5Stack
from transformers.utils.model_parallel_utils import assert_device_map, get_device_map


class EncT5ForSequenceClassification(T5PreTrainedModel):
    _keys_to_ignore_on_load_missing = [
        r"encoder\.embed_tokens\.weight",
    ]

    def __init__(self, config: T5Config, dropout=0.1):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.config = config

        self.shared = nn.Embedding(config.vocab_size, config.d_model)

        encoder_config = copy.deepcopy(config)
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False
        self.encoder = T5Stack(encoder_config, self.shared)

        self.dropout = nn.Dropout(dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def parallelize(self, device_map=None):
        self.device_map = (
            get_device_map(len(self.encoder.block), range(torch.cuda.device_count()))
            if device_map is None
            else device_map
        )
        assert_device_map(self.device_map, len(self.encoder.block))
        self.encoder.parallelize(self.device_map)
        self.classifier = self.classifier.to(self.encoder.first_device)
        self.model_parallel = True

    def deparallelize(self):
        self.encoder.deparallelize()
        self.encoder = self.encoder.to("cpu")
        self.model_parallel = False
        self.device_map = None
        torch.cuda.empty_cache()

    def get_input_embeddings(self):
        return self.shared

    def set_input_embeddings(self, new_embeddings):
        self.shared = new_embeddings
        self.encoder.set_input_embeddings(new_embeddings)

    def get_encoder(self):
        return self.encoder

    def _prune_heads(self, heads_to_prune):
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            head_mask=None,
            inputs_embeds=None,
            labels=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
    ):
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.encoder(
            input_ids=input_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = outputs[0]
        pooled_output = hidden_states[:, 0, :]  # Take bos token (equiv. to <s>)

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        loss = None
        if labels is not None:
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)
        if not return_dict:
            output = (logits,) + outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



import os.path
import pathlib
import shutil
import tempfile

import chromadb
import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.retrieval import bm25, vectordb, hybrid_rrf, hybrid_cc
from autorag.nodes.retrieval.run import run_retrieval_node, select_result_for_hybrid, get_ids_and_scores
from autorag.nodes.retrieval.vectordb import vectordb_ingest
from autorag.utils.util import load_summary_file

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resources_dir = os.path.join(root_dir, "resources")


@pytest.fixture
def node_line_dir():
    with tempfile.TemporaryDirectory() as test_project_dir:
        sample_project_dir = os.path.join(resources_dir, "sample_project")
        # copy & paste all folders and files in sample_project folder
        shutil.copytree(sample_project_dir, test_project_dir, dirs_exist_ok=True)

        chroma_path = os.path.join(test_project_dir, "resources", "chroma")
        os.makedirs(chroma_path)
        db = chromadb.PersistentClient(path=chroma_path)
        collection = db.create_collection(name="openai", metadata={"hnsw:space": "cosine"})
        corpus_path = os.path.join(test_project_dir, "data", "corpus.parquet")
        corpus_df = pd.read_parquet(corpus_path)
        vectordb_ingest(collection, corpus_df, OpenAIEmbedding())

        test_trail_dir = os.path.join(test_project_dir, "test_trial")
        os.makedirs(test_trail_dir)
        node_line_dir = os.path.join(test_trail_dir, "test_node_line")
        os.makedirs(node_line_dir)
        yield node_line_dir


def test_run_retrieval_node(node_line_dir):
    modules = [bm25, vectordb, hybrid_rrf, hybrid_cc, hybrid_cc]
    module_params = [
        {'top_k': 4, 'bm25_tokenizer': 'gpt2'},
        {'top_k': 4, 'embedding_model': 'openai'},
        {'top_k': 4, 'rrf_k': 2, 'target_modules': ('bm25', 'vectordb')},
        {'top_k': 4, 'target_modules': ('bm25', 'vectordb'), 'weights': (0.3, 0.7)},
        {'top_k': 4, 'target_modules': ('bm25', 'vectordb'), 'weights': (0.5, 0.5)},
    ]
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    strategies = {
        'metrics': ['retrieval_f1', 'retrieval_recall'],
        'strategy': 'normalize_mean',
        'speed_threshold': 5,
    }
    previous_result = pd.read_parquet(qa_path)
    best_result = run_retrieval_node(modules, module_params, previous_result, node_line_dir, strategies)
    assert os.path.exists(os.path.join(node_line_dir, "retrieval"))
    expect_columns = ['qid', 'query', 'retrieval_gt', 'generation_gt',
                      'retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_f1', 'retrieval_recall']
    assert all([expect_column in best_result.columns for expect_column in expect_columns])
    # test summary feature
    summary_path = os.path.join(node_line_dir, "retrieval", "summary.csv")
    bm25_top_k_path = os.path.join(node_line_dir, "retrieval", "0.parquet")
    assert os.path.exists(bm25_top_k_path)
    bm25_top_k_df = pd.read_parquet(bm25_top_k_path)
    assert os.path.exists(summary_path)
    summary_df = load_summary_file(summary_path)
    assert set(summary_df.columns) == {'filename', 'retrieval_f1', 'retrieval_recall',
                                       'module_name', 'module_params', 'execution_time', 'is_best'}
    assert len(summary_df) == 5
    assert summary_df['filename'][0] == "0.parquet"
    assert summary_df['retrieval_f1'][0] == bm25_top_k_df['retrieval_f1'].mean()
    assert summary_df['retrieval_recall'][0] == bm25_top_k_df['retrieval_recall'].mean()
    assert summary_df['module_name'][0] == "bm25"
    assert summary_df['module_params'][0] == {'top_k': 4, 'bm25_tokenizer': 'gpt2'}
    assert summary_df['execution_time'][0] > 0
    # assert average times
    assert summary_df['execution_time'][0] + summary_df['execution_time'][1] == pytest.approx(
        summary_df['execution_time'][2])
    assert summary_df['execution_time'][0] + summary_df['execution_time'][1] == pytest.approx(
        summary_df['execution_time'][3])

    assert summary_df['filename'].nunique() == len(summary_df)
    assert len(summary_df[summary_df['is_best'] == True]) == 1

    # test summary_df hybrid retrieval convert well
    assert all(summary_df['module_params'].apply(lambda x: 'ids' not in x))
    assert all(summary_df['module_params'].apply(lambda x: 'scores' not in x))
    hybrid_summary_df = summary_df[summary_df['module_name'].str.contains('hybrid')]
    assert all(hybrid_summary_df['module_params'].apply(lambda x: 'target_modules' in x))
    assert all(hybrid_summary_df['module_params'].apply(lambda x: 'target_module_params' in x))
    assert all(hybrid_summary_df['module_params'].apply(lambda x: x['target_modules'] == ('bm25', 'vectordb')))
    assert all(hybrid_summary_df['module_params'].apply(
        lambda x: x['target_module_params'] == (
        {'top_k': 4, 'bm25_tokenizer': 'gpt2'}, {'top_k': 4, 'embedding_model': 'openai'})))

    # test the best file is saved properly
    best_filename = summary_df[summary_df['is_best'] == True]['filename'].values[0]
    best_path = os.path.join(node_line_dir, "retrieval", f'best_{best_filename}')
    assert os.path.exists(best_path)
    best_df = pd.read_parquet(best_path)
    assert all([expect_column in best_df.columns for expect_column in expect_columns])


@pytest.fixture
def pseudo_node_dir():
    summary_df = pd.DataFrame({
        'filename': ['0.parquet', '1.parquet', '2.parquet'],
        'module_name': ['bm25', 'vectordb', 'vectordb'],
        'module_params': [
            {'top_k': 3},
            {'top_k': 3, 'embedding_model': 'openai'},
            {'top_k': 3, 'embedding_model': 'huggingface'},
        ],
        'execution_time': [1, 1, 1],
        'retrieval_f1': [0.1, 0.2, 0.3],
        'retrieval_recall': [0.2, 0.55, 0.5],
    })
    bm25_df = pd.DataFrame({
        'query': ['query-1', 'query-2', 'query-3'],
        'retrieved_ids': [['id-1', 'id-2', 'id-3'],
                          ['id-1', 'id-2', 'id-3'],
                          ['id-1', 'id-2', 'id-3']],
        'retrieve_scores': [[0.1, 0.2, 0.3],
                            [0.1, 0.2, 0.3],
                            [0.1, 0.2, 0.3]],
        'retrieval_f1': [0.05, 0.1, 0.15],
        'retrieval_recall': [0.1, 0.275, 0.25],
    })
    vector_openai_df = pd.DataFrame({
        'query': ['query-1', 'query-2', 'query-3'],
        'retrieved_ids': [['id-4', 'id-5', 'id-6'],
                          ['id-4', 'id-5', 'id-6'],
                          ['id-4', 'id-5', 'id-6']],
        'retrieve_scores': [[0.3, 0.4, 0.5],
                            [0.3, 0.4, 0.5],
                            [0.3, 0.4, 0.5]],
        'retrieval_f1': [0.15, 0.2, 0.25],
        'retrieval_recall': [0.3, 0.55, 0.5],
    })
    vector_huggingface_df = pd.DataFrame({
        'query': ['query-1', 'query-2', 'query-3'],
        'retrieved_ids': [['id-7', 'id-8', 'id-9'],
                          ['id-7', 'id-8', 'id-9'],
                          ['id-7', 'id-8', 'id-9']],
        'retrieve_scores': [[0.5, 0.6, 0.7],
                            [0.5, 0.6, 0.7],
                            [0.5, 0.6, 0.7]],
        'retrieval_f1': [0.25, 0.3, 0.35],
        'retrieval_recall': [0.5, 0.825, 0.75],
    })

    with tempfile.TemporaryDirectory() as node_dir:
        summary_df.to_csv(os.path.join(node_dir, "summary.csv"))
        bm25_df.to_parquet(os.path.join(node_dir, "0.parquet"))
        vector_openai_df.to_parquet(os.path.join(node_dir, "1.parquet"))
        vector_huggingface_df.to_parquet(os.path.join(node_dir, "2.parquet"))
        yield node_dir


def test_select_result_for_hybrid(pseudo_node_dir):
    filenames = select_result_for_hybrid(pseudo_node_dir, ("bm25", "vectordb"))
    dict_id_scores = get_ids_and_scores(pseudo_node_dir, filenames)
    ids = dict_id_scores['ids']
    scores = dict_id_scores['scores']
    assert len(ids) == len(scores) == 2
    assert len(ids[0]) == len(scores[0]) == 3
    assert len(ids[1]) == len(scores[1]) == 3
    assert ids[0] == [['id-1', 'id-2', 'id-3'],
                      ['id-1', 'id-2', 'id-3'],
                      ['id-1', 'id-2', 'id-3']]
    assert scores[0] == [[0.1, 0.2, 0.3],
                         [0.1, 0.2, 0.3],
                         [0.1, 0.2, 0.3]]
    assert ids[1] == [['id-7', 'id-8', 'id-9'],
                      ['id-7', 'id-8', 'id-9'],
                      ['id-7', 'id-8', 'id-9']]
    assert scores[1] == [[0.5, 0.6, 0.7],
                         [0.5, 0.6, 0.7],
                         [0.5, 0.6, 0.7]]


def test_run_retrieval_node_only_hybrid(node_line_dir):
    modules = [hybrid_cc]
    module_params = [
        {'top_k': 4, 'target_modules': ('bm25', 'vectordb'), 'weights': (0.3, 0.7),
         'target_module_params': ({'top_k': 3}, {'top_k': 3, 'embedding_model': 'openai'})},
    ]
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_path = os.path.join(project_dir, "data", "qa.parquet")
    strategies = {
        'metrics': ['retrieval_f1', 'retrieval_recall'],
    }
    previous_result = pd.read_parquet(qa_path)
    best_result = run_retrieval_node(modules, module_params, previous_result, node_line_dir, strategies)
    assert os.path.exists(os.path.join(node_line_dir, "retrieval"))
    expect_columns = ['qid', 'query', 'retrieval_gt', 'generation_gt',
                      'retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_f1', 'retrieval_recall']
    assert all([expect_column in best_result.columns for expect_column in expect_columns])
    # test summary feature
    summary_path = os.path.join(node_line_dir, "retrieval", "summary.csv")
    single_result_path = os.path.join(node_line_dir, "retrieval", "0.parquet")
    assert os.path.exists(single_result_path)
    single_result_df = pd.read_parquet(single_result_path)
    assert os.path.exists(summary_path)
    summary_df = load_summary_file(summary_path)
    assert set(summary_df.columns) == {'filename', 'retrieval_f1', 'retrieval_recall',
                                       'module_name', 'module_params', 'execution_time', 'is_best'}
    assert len(summary_df) == 1
    assert summary_df['filename'][0] == "0.parquet"
    assert summary_df['retrieval_f1'][0] == single_result_df['retrieval_f1'].mean()
    assert summary_df['retrieval_recall'][0] == single_result_df['retrieval_recall'].mean()
    assert summary_df['module_name'][0] == "hybrid_cc"
    assert summary_df['module_params'][0] == {'top_k': 4, 'target_modules': ('bm25', 'vectordb'), 'weights': (0.3, 0.7),
         'target_module_params': ({'top_k': 3}, {'top_k': 3, 'embedding_model': 'openai'})}
    assert summary_df['execution_time'][0] > 0
    assert summary_df['is_best'][0] == True
    assert summary_df['filename'].nunique() == len(summary_df)

    # test the best file is saved properly
    best_filename = summary_df[summary_df['is_best'] == True]['filename'].values[0]
    best_path = os.path.join(node_line_dir, "retrieval", f'best_{best_filename}')
    assert os.path.exists(best_path)
    best_df = pd.read_parquet(best_path)
    assert all([expect_column in best_df.columns for expect_column in expect_columns])
